/* License of this file: CC0 Public Domain Dedication along
   see http://creativecommons.org/publicdomain/zero/1.0/ */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>

#include <libmbtk/libmbtk.h>
#include <libmbtk/backend/sdl2.h>
#include <libmbtk/widget.h>
#include <libmbtk/timer.h>
#include <libmbtk/util.h>
#include <libmbtk/style_f3d.h>

static char *font[] = {"../../../font/Vera.ttf,size=15"};
static mbtk_backend_t be;
static mbtk_display_t disp;

/* Initialize be and disp and set up f3d for default style. Exits on error. */
static void example_init_sdl2(int *argc, char **argv[])
{
	if (mbtk_sdl2_backend(&be) != 0) {
		fprintf(stderr, "Unable to set up the sdl2 backend\n");
		exit(1);
	}

	if (mbtk_be_init(&be) != 0) {
		fprintf(stderr, "Unable to initialize the sdl2 backend\n");
		exit(1);
	}

	if (mbtk_display_init(&be, &disp, argc, argv) != 0) {
		fprintf(stderr, "Unable to initialize mbtk display with the sdl2 backend\n");
		exit(1);
	}

	/* Set default style and font ID for the display */
	mbtk_f3d_style(&disp);
	disp.default_fid = mbtk_resolve_font(&disp, 1, font);
}


static int clock_minsize(mbtk_canvas_t *c, long *mx, long *my)
{
	*mx = *my = 150;
	return 0;
}

static void *clock_ck_text;

static int clock_draw(mbtk_canvas_t *c, long x_offs, long y_offs, long w, long h)
{
	mbtk_window_t *win = c->subwin;
	long cx = w/2, cy = h/2;
	float size = MBTK_MIN(w, h);
	float r_dial = size * 0.45; /* dial outer radius */
	float l_dial = size * 0.05; /* dial long marking length */
	float arm_len[3];
	int n;
	time_t day_second;
	double ang, step = M_PI*2.0/60.0, arm[3];
	char txt[32];

	day_second = *(time_t *)c->w.user_data;
	arm[0] = fmod((double)day_second, 60.0);          arm_len[0] = r_dial - l_dial;     /* second */
	arm[1] = fmod((double)day_second / 60.0, 60.0);   arm_len[1] = r_dial - 2.0*l_dial; /* minute */
	arm[2] = fmod((double)day_second / 3600.0, 12.0); arm_len[2] = r_dial - 4.0*l_dial; /* hour */

	sprintf(txt, "%02.0f:%02.0f:%02.0f", arm[2], arm[1], arm[0]);
	arm[2] *= 5;

	/* draw background rectangle */
	mbtk_draw_color(win, 255, 255, 200, 255);
	mbtk_draw_fillrect(win, 0, 0, w, h);

	/* draw dial */
	mbtk_draw_color(win, 255, 128, 128, 255);
	for(n = 0, ang = 0; n < 60; n++, ang += step) {
		float cs = cos(ang), sn = sin(ang), len;

		if ((n % 15) == 0) len = l_dial;
		else if ((n % 5) == 0) len = l_dial/2.0;
		else len = l_dial/4.0;

		mbtk_draw_line(win,
			cx + (r_dial - len) * cs, cy + (r_dial - len) * sn,
			cx + r_dial * cs,         cy + r_dial * sn);
	}

	/* draw 3 arms */
	for(n = 0; n < 3; n++) {
		float cs, sn;

		ang = step * arm[n] - M_PI/2.0;
		cs = cos(ang);
		sn = sin(ang);

		switch(n) {
			case 0: mbtk_draw_color(win, 255, 64, 64, 128); break;
			case 1: mbtk_draw_color(win, 64, 255, 64, 128); break;
			case 2: mbtk_draw_color(win, 64, 64, 255, 128); break;
		}

		mbtk_draw_line(win, cx, cy, cx + arm_len[n] * cs, cy + arm_len[n] * sn);
	}

	mbtk_draw_free_text(win, &clock_ck_text);
	mbtk_draw_text(win, cx, cy, txt, 8, &clock_ck_text);

	return 0;
}

static mbtk_event_handled_t clock_ev(mbtk_widget_t *w, mbtk_kw_t id, mbtk_event_t *ev, void *user_data)
{
	if (ev->type == MBTK_EV_MOUSE_PRESS) {
		/* coordinates are relative to the canvas, top right corner being 0;0 */
		printf("Mouse press: %ld %ld button %d\n", ev->data.mouse.x, ev->data.mouse.y, ev->data.mouse.button);
	}
	return MBTK_EVENT_HANDLED;
}

static mbtk_event_handled_t clock_step_bwd_cb(mbtk_widget_t *w, mbtk_kw_t id, void *user_data)
{
	time_t *day_second = user_data;
	*day_second -= 30;
	return 0;
}

static mbtk_event_handled_t clock_step_fwd_cb(mbtk_widget_t *w, mbtk_kw_t id, void *user_data)
{
	time_t *day_second = user_data;
	*day_second += 30;
	return 0;
}


static mbtk_timer_res_t tick_cb(long repleft, void *user_data)
{
	mbtk_canvas_t *clk = user_data;
	time_t *day_second = clk->w.user_data;

	(*day_second)++;
	mbtk_widget_needs_redraw(&clk->w);
	return MBTK_TIMER_CONTINUE;
}

int main(int argc, char *argv[])
{
	mbtk_window_t *win;
	mbtk_box_t *root, *ctrl;
	mbtk_canvas_t *clk;
	mbtk_button_t *fwd, *bwd;
	time_t day_second = time(NULL);

	example_init_sdl2(&argc, &argv);

	/* create a small, normal (non-modal, non-menu) window. -1 for coord means
	   we do not have a request on placement coords; the last two are recommended
	   initial window size */
	win = mbtk_window_new(&disp, MBTK_WNTY_NORMAL, "miniboxtk hello world", -1, -1, 150, 50);
	if (win == NULL) {
		fprintf(stderr, "Unable to create top window\n");
		exit(1);
	}

	/* create the root widget as a vbox so child widgets are packed vertically */
	root = mbtk_vbox_new(NULL);
	mbtk_window_set_root_widget(win, &root->w);


	/*** create canvas for the clock ***/
	clk = mbtk_canvas_new(NULL);
	clk->w.user_data = &day_second;
	mbtk_canvas_set_minsize_cb(clk, clock_minsize);
	mbtk_widget_set_span(&clk->w, HVBOX_FILL); /* take all extra space available */
	mbtk_canvas_set_draw_cb(clk, clock_draw);
	mbtk_box_add_widget(root, &clk->w, 0);
	mbtk_canvas_callback_event(clk, clock_ev, NULL);


	/* create an add a control buttons in the ctrl hbox */
	ctrl = mbtk_hbox_new(NULL);
	mbtk_box_add_widget(root, &ctrl->w, 0);

	bwd = mbtk_button_new_with_label("back");
	mbtk_button_callback_click(bwd, clock_step_bwd_cb, &day_second);
	mbtk_box_add_widget(ctrl, &bwd->w, 0);

	fwd = mbtk_button_new_with_label("forward");
	mbtk_button_callback_click(fwd, clock_step_fwd_cb, &day_second);
	mbtk_box_add_widget(ctrl, &fwd->w, 0);

	/* Increase day_second once a second in tick_cb and redraw clk */
	mbtk_timer_new(&disp, MBTK_TIMER_REPEAT_FOREVER, 1.0, tick_cb, clk);

	/* Main loop: run the GUI until win is closed */
	mbtk_run_window(win);

	/* Uninit everyhting at once */
	mbtk_uninit();
}
