/* License of this file: CC0 Public Domain Dedication along
   see http://creativecommons.org/publicdomain/zero/1.0/ */

#include <stdio.h>
#include <stdlib.h>

#include <libmbtk/libmbtk.h>
#include <libmbtk/backend/sdl2.h>
#include <libmbtk/widget.h>
#include <libmbtk/style_f3d.h>

static char *font[] = {"../../../font/Vera.ttf,size=15"};
static mbtk_backend_t be;
static mbtk_display_t disp;

/* Initialize be and disp and set up f3d for default style. Exits on error. */
static void example_init_sdl2(int *argc, char **argv[])
{
	if (mbtk_sdl2_backend(&be) != 0) {
		fprintf(stderr, "Unable to set up the sdl2 backend\n");
		exit(1);
	}

	if (mbtk_be_init(&be) != 0) {
		fprintf(stderr, "Unable to initialize the sdl2 backend\n");
		exit(1);
	}

	if (mbtk_display_init(&be, &disp, argc, argv) != 0) {
		fprintf(stderr, "Unable to initialize mbtk display with the sdl2 backend\n");
		exit(1);
	}

	/* Set default style and font ID for the display */
	mbtk_f3d_style(&disp);
	disp.default_fid = mbtk_resolve_font(&disp, 1, font);
}

static mbtk_event_handled_t title_clk(mbtk_widget_t *w, mbtk_kw_t id, void *user_data)
{
	mbtk_rim_t *rim = user_data;
	mbtk_widget_t *prev;
	static int cnt = 0;
	char tmp[32];

	sprintf(tmp, "title #%d", cnt++);
	prev = mbtk_rim_set_title_label(rim, tmp);
	if (prev != NULL)
		mbtk_widget_destroy(prev);

	return 0;
}

static mbtk_event_handled_t content_clk(mbtk_widget_t *w, mbtk_kw_t id, void *user_data)
{
	mbtk_box_t *vbox = user_data;
	static int cnt = 0;
	char tmp[32];
	mbtk_label_t *lab;

	sprintf(tmp, "content #%d", cnt++);
	lab = mbtk_label_new(NULL, tmp);
	mbtk_box_add_widget(vbox, &lab->w, 0);

	return 0;
}


int main(int argc, char *argv[])
{
	mbtk_window_t *win;
	mbtk_box_t *root, *vbox;
	mbtk_label_t *lab;
	mbtk_rim_t *rim;
	mbtk_button_t *btn;

	example_init_sdl2(&argc, &argv);

	/* create a small, normal (non-modal, non-menu) window. -1 for coord means
	   we do not have a request on placement coords; the last two are recommended
	   initial window size */
	win = mbtk_window_new(&disp, MBTK_WNTY_NORMAL, "miniboxtk hello world", -1, -1, 200, 200);
	if (win == NULL) {
		fprintf(stderr, "Unable to create top window\n");
		exit(1);
	}

	/* create the root widget as a vbox so child widgets are packed vertically;
	   default alignment is "justify", which means excess space are put between
	   the widgets. */
	root = mbtk_vbox_new(NULL);
	mbtk_window_set_root_widget(win, &root->w);
	mbtk_box_set_padding(root, 15);

	/* Create and add a label */
	lab = mbtk_label_new(NULL, "A rim follows:");
	mbtk_box_add_widget(root, &lab->w, 0);

	/* Create and add a rim with a title string */
	rim = mbtk_rim_new_with_label("Title!");
	mbtk_box_add_widget(root, &rim->w, 0);

	/* Create a vbox for hosting content of the rim */
	vbox = mbtk_vbox_new(NULL);
	mbtk_box_set_span(vbox, HVBOX_FILL);
	mbtk_rim_set_child(rim, &vbox->w);

	/* Create a button and add it in the rim */
	btn = mbtk_button_new_with_label("Change title to label");
	mbtk_button_callback_click(btn, title_clk, rim);
	mbtk_box_add_widget(vbox, &btn->w, 0);

	/* Create a button and add it in the rim */
	btn = mbtk_button_new_with_label("Add content");
	mbtk_button_callback_click(btn, content_clk, vbox);
	mbtk_box_add_widget(vbox, &btn->w, 0);
	mbtk_rim_set_span(rim, HVBOX_FILL);

	/* background color of the rim - does not affect background color of
	   the content vbox, only the titlebox and the frame */
/*	mbtk_rim_style_set_prop(rim, mbtk_kw("bg.color"), mbtk_arg_str("#00ff00"));*/

	/* To set the rim's thicker: */
/*	mbtk_rim_style_set_prop(rim, mbtk_kw("frame.width"), mbtk_arg_long(2));*/

	/* Set minimum height of the title bar of the rim and paint it red */
	if (0) {
		mbtk_box_t *tbar = (mbtk_box_t *)mbtk_rim_subwidget(rim, mbtk_kw("tbar"));
		if (tbar != NULL) {
			mbtk_box_set_minsize(tbar, 0, 30);
			mbtk_box_style_set_prop(tbar, mbtk_kw("bg.color"), mbtk_arg_str("#ff0000"));
			tbar->no_background = 0; /* enable drawing the background fill */
		}
		else
			fprintf(stderr, "ERROR: rim failed to return its tbar widget");
	}

	/* main loop: run the GUI until win is closed */
	mbtk_run_window(win);

	/* Uninit everyhting at once */
	mbtk_uninit();
}
