/*  miniboxtk - small GUI toolkit/widget system
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    (Supported by NLnet NGI0 PET Fund in 2022)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://repo.hu/projects/miniboxtk
      contact lead developer: http://www.repo.hu/projects/miniboxtk/contact.html
*/

#include <stdio.h>
#include <string.h>
#include <assert.h>

static int mbtk_glfw_draw_color(mbtk_window_t *win, int r, int g, int b, int a)
{
	glfw_win_t *gw = win2gw(win);
	gw->red = (float)r/255.0;
	gw->green = (float)g/255.0;
	gw->blue = (float)b/255.0;
	gw->alpha = (float)a/255.0;
	return 0;
}

static void mbtk_glfw_draw_fillrect(mbtk_window_t *win, int x1, int y1, int x2, int y2)
{
	glfw_win_t *gw = win2gw(win);

	primbuf_add(gw, GL_TRIANGLE_FAN, 4);
	vertbuf_add(gw, x1-0.5+1, y1-0.5+1);
	vertbuf_add(gw, x2+0.5+1, y1-0.5+1);
	vertbuf_add(gw, x2+0.5+1, y2+0.5+1);
	vertbuf_add(gw, x1-0.5+1, y2+0.5+1);
}

typedef struct {
	int width, height;
	unsigned has_alpha:1;
	unsigned char pixels[1]; /* overallocated to host the whole pixmap */
} glfw_raw_pixmap_t;

static int mbtk_glfw_draw_import_pixmap(mbtk_display_t *display, mbtk_pixmap_t *dst, unsigned char *pixels, int width, int height, int has_alpha)
{
	glfw_raw_pixmap_t *rp;
	int dx = has_alpha ? 4 : 3;
	long len = width*height*dx;

	dst->bedata.p = NULL;

	rp = malloc(sizeof(glfw_raw_pixmap_t) - 1 + len);
	if (rp == NULL)
		return -1;

	dst->bedata.p = rp;
	rp->width = width;
	rp->height = height;
	rp->has_alpha = has_alpha;
	memcpy(rp->pixels, pixels, len);

	return 0;
}

static void mbtk_glfw_draw_free_pixmap(mbtk_display_t *display, mbtk_pixmap_t *px)
{
	glfw_raw_pixmap_t *rp = px->bedata.p;
	free(rp);
	px->bedata.p = NULL;
	return;
}

static mbtk_px_id_t mbtk_glfw_draw_bind_pixmap(mbtk_window_t *win, mbtk_pixmap_t *px)
{
	mbtk_px_id_t pxid;
	glfw_raw_pixmap_t *rp = px->bedata.p;

	if (rp != NULL) {
		pxid.p = rp;
		pxid.l = texture_import(rp->pixels, rp->width, rp->height, rp->has_alpha);
	}
	else {
		pxid.p = NULL;
		pxid.l = -1;
	}

	return pxid;
}

static void mbtk_glfw_draw_unbind_pixmap(mbtk_window_t *win, mbtk_px_id_t pxid)
{
	if (pxid.l != -1)
		texture_free(pxid.l);
}


static void mbtk_glfw_draw_pixmap(mbtk_window_t *win, int x, int y, mbtk_px_id_t pxid)
{
	glfw_win_t *gw = win2gw(win);
	mbtk_pixmap_t *px = pxid.p;
	prim_t *p;
	GLfloat xw, yh;
	GLfloat x1, y1, x2, y2, x3, y3, x4, y4;
	GLfloat u1, v1, u2, v2, u3, v3, u4, v4;

	xw = x + px->w; yh = y + px->h; /* opposite corner */

	x1 = x;    y1 = y;    u1 = 0.0; v1 = 0.0;
	x2 = xw;   y2 = y;    u2 = 1.0; v2 = 0.0;
	x3 = xw;   y3 = yh;   u3 = 1.0; v3 = 1.0;
	x4 = x;    y4 = yh;   u4 = 0.0; v4 = 1.0;

	p = primbuf_add(gw, GL_TRIANGLE_FAN, 4);
	vertbuf_add_xyuv(gw, x1, y1, u1, v1);
	vertbuf_add_xyuv(gw, x2, y2, u2, v2);
	vertbuf_add_xyuv(gw, x3, y3, u3, v3);
	vertbuf_add_xyuv(gw, x4, y4, u4, v4);
	p->txtid = pxid.l;
}

static void mbtk_glfw_draw_line_(mbtk_window_t *win, GLfloat x1, GLfloat y1, GLfloat x2, GLfloat y2)
{
	glfw_win_t *gw = win2gw(win);

	primbuf_add(gw, GL_LINES, 2);
	vertbuf_add(gw, x1+1.0, y1+1.0);
	vertbuf_add(gw, x2+1.0, y2+1.0);
}

static void mbtk_glfw_draw_line(mbtk_window_t *win, int x1, int y1, int x2, int y2)
{
	if (x1 < x2) x1--;
	else if (x2 < x1) x2--;

	if (y1 < y2) y1--;
	else if (y2 < y1) y2--;

	mbtk_glfw_draw_line_(win, x1, y1, x2, y2);
}

static void mbtk_glfw_draw_hline(mbtk_window_t *win, int x1, int y1, int x2)
{
	if (x2 > x1)
		mbtk_glfw_draw_line_(win, x1-0.5, y1, x2+0.5, y1);
	else
		mbtk_glfw_draw_line_(win, x1+0.5, y1, x2+0.5, y1);
}

static void mbtk_glfw_draw_vline(mbtk_window_t *win, int x1, int y1, int y2)
{
	if (y2 > y1)
		mbtk_glfw_draw_line_(win, x1, y1-0.5, x1, y2+0.5);
	else
		mbtk_glfw_draw_line_(win, x1, y1+0.5, x1, y2-0.5);
}

static int mbtk_glfw_draw_set_font(mbtk_window_t *win, mbtk_font_id_t fontid)
{
	glfw_win_t *gw = win2gw(win);

	if ((fontid.l == -1) || (fontid.p == NULL))
		return -1;

	gw->font = fontid.p;
	return 0;
}

static void mbtk_glfw_draw_query_text(mbtk_window_t *win, int *w, int *ascent, int *descent, char *s, int slen)
{
	glfw_win_t *gw = win2gw(win);
	int x;
	FT_GlyphSlot slot;

	if (gw->font == NULL) {
		*ascent = *descent = *w = 0;
		return;
	}

	*ascent = gw->font->asc;
	*descent = gw->font->dsc;

	slot = gw->font->face->glyph;
	for(x = 0; *s != '\0'; s++) {
		if (FT_Load_Char(gw->font->face, *s, 0) != 0)
			continue;
		x += slot->advance.x;
	}

	*w = x/64;
}

static long mbtk_glfw_draw_query_text_pos(mbtk_window_t *win, long *char_start_px, long *char_end_px, long pxpos, long delta_chr, char *s, int slen)
{
	glfw_win_t *gw = win2gw(win);
	int found = 0;
	long x, x2, lastx, chidx;
	FT_GlyphSlot slot;

	if (gw->font == NULL)
		return -1;

	pxpos *= 64;

	slot = gw->font->face->glyph;
	for(chidx = x = 0; *s != '\0'; s++, x = x2, chidx++) {
		if (FT_Load_Char(gw->font->face, *s, 0) != 0)
			continue;
		x2 = x + slot->advance.x;

		if ((pxpos >= x) && (pxpos < x2))
			found = 1;
		if (found) {
			if (delta_chr <= 0)
				break;
			delta_chr--;
		}
		lastx = x;
	}

	if (!found) {
		if (pxpos < x2)
			return -1;
		/* pxpos is to the right of the text - return last char */
		*char_start_px = x2/64+1;
		*char_end_px = x2/64+2;
		return slen-1;
	}

	if (found) {
		if (delta_chr < 0) {
			if (chidx == 0)
				return -1;
			*char_start_px = lastx/64+1;
			*char_end_px = x/64+1;
			return chidx - 1;
		}

		if (delta_chr > 0)
			return -1; /* beyond the end */

		*char_start_px = x/64+1;
		*char_end_px = x2/64+1;
		return chidx;
	}

	return -1;
}

static void ttf_draw_bitmap(glfw_win_t *gw, unsigned char *dst, int dstw, int dsth, FT_Bitmap *bitmap, FT_Int ox, FT_Int oy, int trx, int try, int ymirror)
{
	FT_Int x, y, p, q, w;
	FT_Int x_max = ox + bitmap->width;
	FT_Int y_max = oy + bitmap->rows;

	assert(bitmap->pixel_mode == FT_PIXEL_MODE_GRAY);

	for(x = ox, p = 0; x < x_max; x++, p++) {
		int xx = x + trx;
		for(y = oy, q = 0; y < y_max; y++, q++) {
			int yy;
			long offs;

			if (ymirror)
				yy = (dsth - y - 1) + try;
			else
				yy = y + try;

			if ((xx < 0)  || (yy < 0) || (xx >= dstw) || (yy >= dsth))
				continue;
			w = q * bitmap->width;
	
			offs = (yy*dstw + xx) * 4;
			dst[offs + 3] |= bitmap->buffer[w + p];
		}
	}
}

typedef struct {
	mbtk_pixmap_t px;   /* doesn't have the actual raw pixmap data, only dimensions */
	int asc;            /* ascent so rendering coords can be transformed */
	mbtk_px_id_t pxid;
} txt_pxmap_t;

static void mbtk_glfw_draw_text(mbtk_window_t *win, int x, int y, char *s, int slen, void **cache)
{
	glfw_win_t *gw = win2gw(win);
	long txtid;
	txt_pxmap_t *tpm;

	if (*cache == NULL) {
		int w, asc, dsc, h, dx, n, tot;
		unsigned char *tmp, *i;
		FT_GlyphSlot slot;

		mbtk_glfw_draw_query_text(win, &w, &asc, &dsc, s, slen);

		h = asc+dsc;
		tot = w*h*4;
		tmp = malloc(tot);

		slot = gw->font->face->glyph;

		for(n = 0, i = tmp; n < tot; n += 4, i+=4) {
			i[0] = gw->red * 255.0;
			i[1] = gw->green * 255.0;
			i[2] = gw->blue * 255.0;
			i[3] = 0;
		}

		for(dx = 0; *s != '\0'; s++) {
			if (FT_Load_Char(gw->font->face, *s, FT_LOAD_RENDER) != 0)
				continue;
			ttf_draw_bitmap(gw, tmp, w, h, &slot->bitmap, slot->bitmap_left, 0, dx, (asc - slot->metrics.horiBearingY/64), 0);
			dx += slot->advance.x / 64;
		}

		txtid = texture_import(tmp, w, h, 1);
		free(tmp);
		if (txtid == 0)
			return;

		tpm = calloc(sizeof(txt_pxmap_t), 1);
		*cache = tpm;

		tpm->asc = asc;
		tpm->px.w = w;
		tpm->px.h = h;
		tpm->px.has_alpha = 1;
		
		tpm->pxid.p = &tpm->px;
		tpm->pxid.l = txtid;
	}
	else
		tpm = *cache;

	mbtk_glfw_draw_pixmap(win, x, y - tpm->asc, tpm->pxid);
}

static void mbtk_glfw_draw_free_text(mbtk_window_t *win, void **cache)
{
	txt_pxmap_t *tpm = *cache;

	if (tpm == NULL)
		return;

	mbtk_glfw_draw_unbind_pixmap(win, tpm->pxid);
	texture_free(tpm->pxid.l);
	free(tpm);
	*cache = NULL;
}

