/*  miniboxtk - small GUI toolkit/widget system
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    (Supported by NLnet NGI0 PET Fund in 2022)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://repo.hu/projects/miniboxtk
      contact lead developer: http://www.repo.hu/projects/miniboxtk/contact.html
*/

static glfw_win_t *mbtk_glfw_get_subw(glfw_win_t *parent_win, long x, long y, long *subx, long *suby)
{
	glfw_win_t *subwin;
	for(subwin = gdl_first(&parent_win->subwins); subwin != NULL; subwin = subwin->link.next) {
		mbtk_window_t *sw = (mbtk_window_t *)subwin;
		if ((x >= sw->x) && (y >= sw->y) && (x < sw->x + sw->w) && (y < sw->y + sw->h))
			return mbtk_glfw_get_subw(subwin, x - sw->x, y - sw->y, subx, suby);
	}
	*subx = x; *suby = y;
	return parent_win;
}

static mbtk_glfw_event_t *alloc_ev(glfw_win_t *gw)
{
	mbtk_window_t *mw = (mbtk_window_t *)gw;
	glfw_display_t *gd = (glfw_display_t *)mw->display;
	mbtk_glfw_event_t *gev = uall_slab_alloc(&gd->evs);

	if (gd->evq_tail != NULL) {
		gd->evq_tail->next = gev;
		gd->evq_tail = gev;
	}
	else
		gd->evq_tail = gd->evq = gev;

	gev->ev.win = mw;
	gev->timestamp = glfwGetTime();
	gev->next = NULL;

	return gev;
}

static void free_ev_first(glfw_display_t *gs) {
	mbtk_glfw_event_t *gev = gs->evq;

	gs->evq = gev->next;
	if (gs->evq == NULL)
		gs->evq_tail = NULL;

	uall_slab_free(&gs->evs, gev);
}

static void mbtk_glfw_ev_kbd(GLFWwindow *window, int key, int scancode, int action, int mods)
{
	glfw_win_t *gw = glfwGetWindowUserPointer(window);
	mbtk_glfw_event_t *gev = alloc_ev(gw);

	gev->ev.type = (action == GLFW_PRESS) ? MBTK_EV_KEY_PRESS : MBTK_EV_KEY_RELEASE;
	gev->ev.data.key.scancode = scancode;
	mbtk_glfw_key_xlate(&gev->ev, key, mods);
	gev->ev.data.key.edit = 0;
}

static void mbtk_glfw_ev_chr(GLFWwindow *window, unsigned int unicode)
{
	glfw_win_t *gw = glfwGetWindowUserPointer(window);
	mbtk_glfw_event_t *gev = alloc_ev(gw);

	gev->ev.type = MBTK_EV_KEY_RELEASE;
	gev->ev.data.key.scancode = 0;
	gev->ev.data.key.sym = 0;
	gev->ev.data.key.mods = 0;
	gev->ev.data.key.edit = unicode;
}

static void mbtk_glfw_ev_mouse_button(GLFWwindow *window, int button, int action, int mods)
{
	glfw_win_t *gw_ = glfwGetWindowUserPointer(window);
	glfw_win_t *gw = gw_->last_mouse_subw != NULL ? gw_->last_mouse_subw : gw_;
	mbtk_window_t *mw = (mbtk_window_t *)gw;
	mbtk_glfw_event_t *gev = alloc_ev(gw);

	gev->ev.type = (action == GLFW_PRESS) ? MBTK_EV_MOUSE_PRESS : MBTK_EV_MOUSE_RELEASE;
	gev->ev.data.mouse.button = mbtk_glfw_button_xlate(button);
	if (gev->ev.win->entered || gw->subwin) {
		gev->ev.data.mouse.x = gw->last_mouse_x + ((mbtk_window_t *)gw)->content_x0;
		gev->ev.data.mouse.y = gw->last_mouse_y + ((mbtk_window_t *)gw)->content_y0;
	}
	else
		gev->ev.data.mouse.x = gev->ev.data.mouse.y = -1;

	mbtk_display_set_mouse_button(mw->display, gev->ev.data.mouse.button, (action == GLFW_PRESS) ? 1 : 0);
}

static void mbtk_glfw_ev_mouse_motion(GLFWwindow *window, double xpos, double ypos)
{
	glfw_win_t *gw, *gw_ = glfwGetWindowUserPointer(window);
	mbtk_glfw_event_t *gev;
	long subx, suby;

	gw = mbtk_glfw_get_subw(gw_, xpos, ypos, &subx, &suby);
	gev = alloc_ev(gw);
	gev->ev.type = MBTK_EV_MOUSE_MOTION;
	gev->ev.data.mouse.button = 0;

	gw_->last_mouse_x = xpos;
	gw_->last_mouse_y = ypos;
	gw->last_mouse_x = subx;
	gw->last_mouse_y = suby;
	gev->ev.data.mouse.x = subx + ((mbtk_window_t *)gw)->content_x0;
	gev->ev.data.mouse.y = suby + ((mbtk_window_t *)gw)->content_y0;

	gw_->last_mouse_subw = gw;
}

static void mbtk_glfw_ev_mouse_scroll(GLFWwindow *window, double xoffset, double yoffset)
{
	glfw_win_t *gw_ = glfwGetWindowUserPointer(window);
	glfw_win_t *gw = gw_->last_mouse_subw != NULL ? gw_->last_mouse_subw : gw_;
	mbtk_glfw_event_t *gev = alloc_ev(gw);

	gev->ev.type = MBTK_EV_MOUSE_SCROLL;
	gev->ev.data.mouse.button = 0;
	gev->ev.data.mouse.x = xoffset;
	gev->ev.data.mouse.y = yoffset;
}

static void mbtk_glfw_ev_wrefr(GLFWwindow *window)
{
	glfw_win_t *gw = glfwGetWindowUserPointer(window);
	mbtk_glfw_event_t *gev = alloc_ev(gw);

	gev->ev.type = MBTK_EV_WIN_EXPOSE;
	gev->ev.data.win.x = gev->ev.data.win.y = gev->ev.data.win.w = gev->ev.data.win.h = 0;
}

static void mbtk_glfw_ev_wpos(GLFWwindow *window, int x, int y)
{
	glfw_win_t *gw = glfwGetWindowUserPointer(window);
	mbtk_glfw_event_t *gev = alloc_ev(gw);

	gev->ev.type = MBTK_EV_WIN_MOVE;
	gev->ev.data.win.x = x;
	gev->ev.data.win.y = y;
	gev->ev.data.win.w = gev->ev.data.win.h = 0;
}

static void mbtk_glfw_ev_wresize(GLFWwindow *window, int width, int height)
{
	glfw_win_t *gw = glfwGetWindowUserPointer(window);
	mbtk_glfw_event_t *gev = alloc_ev(gw);

	gev->ev.type = MBTK_EV_WIN_RESIZE;
	gev->ev.data.win.x = gev->ev.data.win.y = 0;
	gev->ev.data.win.w = width;
	gev->ev.data.win.h = height;
}

static void mbtk_glfw_ev_wclose(GLFWwindow *window)
{
	glfw_win_t *gw = glfwGetWindowUserPointer(window);
	mbtk_glfw_event_t *gev = alloc_ev(gw);

	gev->ev.type = MBTK_EV_WIN_CLOSE;
	gev->ev.data.win.x = gev->ev.data.win.y = gev->ev.data.win.w = gev->ev.data.win.h = 0;
}


static void mbtk_glfw_ev_wenter(GLFWwindow *window, int entering)
{
	glfw_win_t *gw = glfwGetWindowUserPointer(window);
	mbtk_glfw_event_t *gev = alloc_ev(gw);

	gev->ev.type = entering ? MBTK_EV_WIN_ENTER : MBTK_EV_WIN_LEAVE;
	gev->ev.data.win.x = gev->ev.data.win.y = gev->ev.data.win.w = gev->ev.data.win.h = 0;
}


static void mbtk_glfw_win_event_setup(glfw_win_t *gw)
{
	glfwSetKeyCallback(gw->window, mbtk_glfw_ev_kbd);
	glfwSetCharCallback(gw->window, mbtk_glfw_ev_chr);
	glfwSetMouseButtonCallback(gw->window, mbtk_glfw_ev_mouse_button);
	glfwSetScrollCallback(gw->window, mbtk_glfw_ev_mouse_scroll);
	glfwSetCursorPosCallback(gw->window, mbtk_glfw_ev_mouse_motion);
	glfwSetWindowRefreshCallback(gw->window, mbtk_glfw_ev_wrefr);
	glfwSetWindowPosCallback(gw->window, mbtk_glfw_ev_wpos);
	glfwSetWindowSizeCallback(gw->window, mbtk_glfw_ev_wresize);
	glfwSetWindowCloseCallback(gw->window, mbtk_glfw_ev_wclose);
	glfwSetCursorEnterCallback(gw->window, mbtk_glfw_ev_wenter);
}
