/*  miniboxtk - small GUI toolkit/widget system
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    (Supported by NLnet NGI0 PET Fund in 2022)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://repo.hu/projects/miniboxtk
      contact lead developer: http://www.repo.hu/projects/miniboxtk/contact.html
*/

#define INCLUDE_SELECT
#include "../../config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <libmbtk/src_3rd/genht/hash.h>
#include "../../libmbtk.h"
#include "../common/fdwatch.h"

#include "../test.h"

#include "mbtk_test_priv.h"

#include "draw.c"
#include "iterate.c"
#include "window.c"

static int mbtk_test_display_init(mbtk_backend_t *be, mbtk_display_t *disp, int *argc, char **argv[])
{
	test_display_t *td = (test_display_t *)disp;

	td->clipb = NULL;
	td->win_ids = td->pxids = 0;
	mbtk_be_check_priv_sizes(test_display_t, test_win_t);

	memset(&td->watch, 0, sizeof(td->watch));
	mbtk_watch_fd_set(&td->watch, 0, 1, 0); /* read stdin for commands */

	printf("display_init\n");

	return 0;
}

static int mbtk_test_display_uninit(mbtk_backend_t *be, mbtk_display_t *disp)
{
	test_display_t *td = (test_display_t *)disp;

	mbtk_watch_fd_uninit(&td->watch);

	free(td->clipb);
	printf("display_uninit\n");
	return 0;
}

static double mbtk_test_get_time_el(mbtk_display_t *disp)
{
	test_display_t *td = (test_display_t *)disp;
	printf("get_time_el: -> %f\n", td->time_el);
	return td->time_el;
}

static mbtk_keymod_t mbtk_test_query_kbd_mods(mbtk_display_t *disp)
{
	test_display_t *td = (test_display_t *)disp;
	return td->keymods;
}

static void mbtk_test_clipboard_set(mbtk_display_t *disp, const char *new_content)
{
	test_display_t *td = (test_display_t *)disp;

	free(td->clipb);
	td->clipb = mbtk_strdup(new_content);
	printf("clipboard_set: '%s'\n", td->clipb);
}

static char *mbtk_test_clipboard_get(mbtk_display_t *disp)
{
	test_display_t *td = (test_display_t *)disp;

	printf("clipboard_get: -> '%s'\n", td->clipb);
	return td->clipb;
}

static void mbtk_test_clipboard_free(mbtk_display_t *disp, char *str_got)
{
	printf("clipboard_free: '%s'\n", str_got);
	*str_got = '\0';
	free(str_got);
}

static mbtk_font_id_t mbtk_test_resolve_font(mbtk_display_t *display, int argc, char *argv[])
{
	mbtk_font_id_t idx;

	printf("resolve font\n");

	idx.l = 1;
	return idx;
}

struct mbtk_mocu_s {
	int dummy;
};

static mbtk_mocu_t *mbtk_test_mocu_create_named(mbtk_display_t *disp, mbtk_mocu_named_t name)
{
	static mbtk_mocu_t mocu;
	printf("mocu_create_named: %d\n", name);
	return &mocu;
}

static mbtk_mocu_t *mbtk_test_mocu_create_pixmap(mbtk_display_t *disp, int width, int height, unsigned const char *px, unsigned const char *mask, int hot_x, int hot_y)
{
	static mbtk_mocu_t mocu;
	printf("mocu_create_pixmap\n");
	return &mocu;
}

static void mbtk_test_mocu_free_cursor(mbtk_display_t *disp, mbtk_mocu_t *curs)
{
	printf("mocu_free_cursor\n");
}

static void mbtk_test_mocu_set_cursor(mbtk_display_t *disp, mbtk_mocu_t *curs)
{
	printf("mocu_set_cursor\n");
}

static void mbtk_test_watch_stdin_read(mbtk_display_t *disp, int enable)
{
	printf("watch_stdin_read: %d\n", enable);
}

static void mbtk_test_watch_fd(mbtk_display_t *disp, int fd, int for_read, int for_write)
{
	printf("watch_fd: %d, %d, %d\n", fd, for_read, for_write);
}

int mbtk_test_backend(mbtk_backend_t *dst)
{
	memset(dst, 0, sizeof(mbtk_backend_t));

	dst->name = "test";

	dst->display_init = mbtk_test_display_init;
	dst->display_uninit = mbtk_test_display_uninit;
	dst->window_new = mbtk_test_window_new;
	dst->window_close = mbtk_test_window_close;
	dst->win2render_win = mbtk_test_win2render_win;
	dst->clipboard_set = mbtk_test_clipboard_set;
	dst->clipboard_get = mbtk_test_clipboard_get;
	dst->clipboard_free = mbtk_test_clipboard_free;
	dst->iterate = mbtk_test_iterate;
	dst->get_time_el = mbtk_test_get_time_el;

	dst->window_set_title = mbtk_test_window_set_title;
	dst->window_raise = mbtk_test_window_raise;
	dst->window_hide = mbtk_test_window_hide;
	dst->window_show = mbtk_test_window_show;
	dst->window_transient4 = mbtk_test_window_transient4;
	dst->window_set_min_size = mbtk_test_window_set_min_size;
	dst->window_set_size = mbtk_test_window_set_size;

	dst->draw_color = mbtk_test_draw_color;
	dst->draw_fillrect = mbtk_test_draw_fillrect;
	dst->draw_line = mbtk_test_draw_line;
	dst->draw_hline = mbtk_test_draw_hline;
	dst->draw_vline = mbtk_test_draw_vline;
	dst->draw_flush = mbtk_test_draw_flush;
	dst->draw_start = mbtk_test_draw_start;
	dst->draw_set_font = mbtk_test_draw_set_font;
	dst->draw_query_text = mbtk_test_draw_query_text;
	dst->draw_query_text_pos = mbtk_test_draw_query_text_pos;
	dst->draw_text = mbtk_test_draw_text;

	dst->resolve_font = mbtk_test_resolve_font;

	dst->draw_import_pixmap = mbtk_test_draw_import_pixmap;
	dst->draw_free_pixmap = mbtk_test_draw_free_pixmap;
	dst->draw_bind_pixmap = mbtk_test_draw_bind_pixmap;
	dst->draw_pixmap = mbtk_test_draw_pixmap;

	dst->mocu_create_named = mbtk_test_mocu_create_named;
	dst->mocu_create_pixmap = mbtk_test_mocu_create_pixmap;
	dst->mocu_free_cursor = mbtk_test_mocu_free_cursor;
	dst->mocu_set_cursor = mbtk_test_mocu_set_cursor;

	dst->query_kbd_mods = mbtk_test_query_kbd_mods;

	dst->watch_stdin_read = mbtk_test_watch_stdin_read;
	dst->watch_fd = mbtk_test_watch_fd;

	dst->has_alpha = 1;

	return 0;
}
