/*  miniboxtk - small GUI toolkit/widget system
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    (Supported by NLnet NGI0 PET Fund in 2022)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://repo.hu/projects/miniboxtk
      contact lead developer: http://www.repo.hu/projects/miniboxtk/contact.html
*/

#include <limits.h>
#if USHRT_MAX == 65535
typedef unsigned short mbtk_uint16_t;
#else
#error Short int is not 16 bits wide.
#endif

#if UINT_MAX == 4294967295U
typedef unsigned int mbtk_uint32_t;
#elif ULONG_MAX == 4294967295UL
typedef unsigned long mbtk_uint32_t;
#else
#error Neither int nor long is 32 bits wide.
#endif

/* mask to shift: find index of righmost 1
   This code uses a 32-bit version of algorithm to find the rightmost
   one bit in Knuth, _The Art of Computer Programming_, volume 4A
   (draft fascicle), section 7.1.3, "Bitwise tricks and techniques." */
static int get_shift(mbtk_uint32_t mask)
{
	mbtk_uint32_t x;
	const mbtk_uint32_t a = 0x05f66a47; /* magic number, found by brute force */
	static const unsigned decode[32] = { 0, 1, 2, 26, 23, 3, 15, 27, 24, 21, 19, 4, 12, 16, 28, 6, 31, 25, 22, 14, 20, 18, 11, 5, 30, 13, 17, 10, 29, 9, 8, 7 };

	x = (a * (mask & (-mask)))/* & 0xffffffff*/;
	return decode[x >> 27];
}

void draw_Z32(XImage *img, unsigned char *pixels, int has_alpha)
{
	int x, y, dx = img->bits_per_pixel >> 3, dy = img->bytes_per_line, sx = has_alpha ? 4:3;
	int rs = get_shift(img->red_mask), gs = get_shift(img->green_mask), bs = get_shift(img->blue_mask);
	char *dr = img->data, *d;
	unsigned char *s = pixels;

	for(y = 0; y < img->height; y++, dr += dy) {
		for(x = 0, d = dr; x < img->width; x++, d += dx, s += sx) {
			unsigned char r = s[0], g = s[1], b = s[2];
			mbtk_uint32_t *p = (mbtk_uint32_t *)d;
			*p = (mbtk_uint32_t)r << rs | (mbtk_uint32_t)g << gs | (mbtk_uint32_t)b << bs;
		}
	}
}

void draw_Z16(XImage *img, unsigned char *pixels, int has_alpha)
{
	int x, y, dx = img->bits_per_pixel >> 3, dy = img->bytes_per_line, sx = has_alpha ? 4:3;
	int rs = get_shift(img->red_mask), gs = get_shift(img->green_mask), bs = get_shift(img->blue_mask);
	int rm = img->red_mask >> rs, gm = img->green_mask >> gs, bm = img->blue_mask >> bs;
	int rd = 255/rm, gd = 255/gm, bd = 255/bm;
	char *dr = img->data, *d;
	unsigned char *s = pixels;

	for(y = 0; y < img->height; y++, dr += dy) {
		for(x = 0, d = dr; x < img->width; x++, d += dx, s += sx) {
			unsigned char r = s[0], g = s[1], b = s[2];
			mbtk_uint16_t *p = (mbtk_uint16_t *)d;
			r = r / rd;
			g = g / gd;
			b = b / bd;
			*p = (mbtk_uint16_t)r << rs | (mbtk_uint16_t)g << gs | (mbtk_uint16_t)b << bs;
		}
	}
}

void draw_Z8(XImage *img, unsigned char *pixels, int has_alpha, Display *display, Colormap cmap)
{
	int x, y, dx = img->bits_per_pixel >> 3, dy = img->bytes_per_line, sx = has_alpha ? 4:3;
	char *dr = img->data, *d;
	unsigned short cache[64];
	unsigned char *s = pixels;

	memset(cache, 0xff, sizeof(cache));

	for(y = 0; y < img->height; y++, dr += dy) {
		for(x = 0, d = dr; x < img->width; x++, d += dx, s += sx) {
			unsigned char r = s[0], g = s[1], b = s[2], id;
			XColor clr;
			id = (r & 0xC0) >> 6 | (g & 0xC0) >> 4 | (b & 0xC0) >> 2;
			if (cache[id] == 0xffff) {
				clr.red =   (r & 0xC0) << 8;
				clr.green = (g & 0xC0) << 8;
				clr.blue =  (b & 0xC0) << 8;
				XAllocColor(display, cmap, &clr);
				cache[id] = clr.pixel;
			}
			*d = cache[id];
		}
	}
}

/* Calculate a clipmasp pixmap so that transparent pixels (alpha == 0) can be
   drawn without proper alpha blending */
void draw_alpha(Display *display, Pixmap pxm, unsigned char *pixels, int w, int h)
{
	int x, y;
	unsigned char *s = pixels;
	GC gc;

	gc = XCreateGC(display, pxm, 0, 0);
	XSetForeground(display, gc, 1);
	XFillRectangle(display, pxm, gc, 0, 0, w, h); /* assume all pixels are visible */
	XSetForeground(display, gc, 0);

	for(y = 0; y < h; y++) {
		for(x = 0; x < w; x++, s += 4) {
			unsigned char a = s[3];
			if (a < 0x40) /* threshold - anything less visible than this is transparent */
				XDrawPoint(display, pxm, gc, x, y); /* clear with fgcolor 0 */
		}
	}

	XFreeGC(display, gc);
}


int create_ximage(mbtk_pixmap_t *dst, Display *display, Drawable drw, Visual *visual, int depth, unsigned char *pixels, int width, int height, int has_alpha)
{
	char *raw;
	XImage *img;
	Pixmap mask = 0;

	if ((depth != 32) && (depth != 24) && (depth != 16) && (depth != 8))
		return -1;

	raw = malloc(width * height * 4);
	img = XCreateImage(display, visual, depth, ZPixmap, 0, raw, width, height, 32, 0);
	if (has_alpha) {
		mask = XCreatePixmap(display, drw, width, height, 1);
		draw_alpha(display, mask, pixels, width, height);
	}

	if ((img->bits_per_pixel == 32) || (img->bits_per_pixel == 24))
		draw_Z32(img, pixels, has_alpha);
	else if (img->bits_per_pixel == 16)
		draw_Z16(img, pixels, has_alpha);
	else if (img->bits_per_pixel == 8)
		draw_Z8(img, pixels, has_alpha, display, XDefaultColormap(display, DefaultScreen(display)));

	dst->bedata.raw = raw;
	dst->bedata.p = img;
	dst->bedata.ul = mask;

	return 0;
}

