/*  miniboxtk - small GUI toolkit/widget system
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    (Supported by NLnet NGI0 PET Fund in 2022)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://repo.hu/projects/miniboxtk
      contact lead developer: http://www.repo.hu/projects/miniboxtk/contact.html
*/

#include <stdlib.h>
#include <genht/htip.h>
#include <genht/htsp.h>
#include <genht/hash.h>
#include "keysym.h"

typedef struct {
	mbtk_keysym_t sym;
	const char *name;
} keyname_t;

static keyname_t keynames[] = {
	{MBTK_KS_BACKSPACE, "Backspace"},
	{MBTK_KS_TAB, "Tab"},
	{MBTK_KS_ENTER, "Enter"},
	{MBTK_KS_ESC, "Esc"},

	{MBTK_KS_FN+1, "F1"},
	{MBTK_KS_FN+2, "F2"},
	{MBTK_KS_FN+3, "F3"},
	{MBTK_KS_FN+4, "F4"},
	{MBTK_KS_FN+5, "F5"},
	{MBTK_KS_FN+6, "F6"},
	{MBTK_KS_FN+7, "F7"},
	{MBTK_KS_FN+8, "F8"},
	{MBTK_KS_FN+9, "F9"},
	{MBTK_KS_FN+10, "F10"},
	{MBTK_KS_FN+11, "F11"},
	{MBTK_KS_FN+12, "F12"},
	{MBTK_KS_FN+13, "F13"},
	{MBTK_KS_FN+14, "F14"},
	{MBTK_KS_FN+15, "F15"},
	{MBTK_KS_FN+16, "F16"},
	{MBTK_KS_FN+17, "F17"},
	{MBTK_KS_FN+18, "F18"},
	{MBTK_KS_FN+19, "F19"},
	{MBTK_KS_FN+20, "F20"},

	{MBTK_KS_HOME, "Home"},
	{MBTK_KS_END, "End"},
	{MBTK_KS_LEFT, "Left"},
	{MBTK_KS_RIGHT, "Right"},
	{MBTK_KS_UP, "Up"},
	{MBTK_KS_DOWN, "Down"},
	{MBTK_KS_PGUP, "PageUp"},
	{MBTK_KS_PGDN, "PageDown"},

	{MBTK_KS_SHIFT_L, "LeftShift"},
	{MBTK_KS_SHIFT_R, "RightShift"},
	{MBTK_KS_ALT_L, "LeftAlt"},
	{MBTK_KS_ALT_R, "RightAlt"},
	{MBTK_KS_CTRL_L, "LeftCtrl"},
	{MBTK_KS_CTRL_R, "RightCtrl"},

	{MBTK_KS_DEL, "Del"},
	{MBTK_KS_INS, "Insert"},
	{0, NULL}
};

#define NUM_KEYSYMS (sizeof(keynames)/sizeof(keynames[0]))

static int keysym2name_inited = 0;
static htip_t keysym2name;

static int name2keysym_inited = 0;
static htsp_t name2keysym;


const char *mbtk_keysym2name(mbtk_keysym_t ks)
{
	keyname_t *n;

	if (!keysym2name_inited) {
		htip_init(&keysym2name, longhash, longkeyeq);
		htip_resize(&keysym2name, NUM_KEYSYMS);

		for(n = keynames; n->name != NULL; n++)
			htip_set(&keysym2name, n->sym, n);

		keysym2name_inited = 1;
	}

	n = htip_get(&keysym2name, ks);
	return (n == NULL) ? NULL : n->name;
}

mbtk_keysym_t mbtk_name2keysym(const char *name)
{
	keyname_t *n;

	if (!name2keysym_inited) {
		htsp_init(&name2keysym, strhash_case, strkeyeq_case);
		htsp_resize(&name2keysym, NUM_KEYSYMS);

		for(n = keynames; n->name != NULL; n++)
			htsp_set(&name2keysym, (char *)n->name, n);

		name2keysym_inited = 1;
	}

	n = htsp_get(&name2keysym, name);
	return (n == NULL) ? MBTK_KS_INVALID : n->sym;
}

void mbtk_keysym_uninit(void)
{
	htsp_uninit(&name2keysym);
	name2keysym_inited = 0;
	htip_uninit(&keysym2name);
	keysym2name_inited = 0;
}

