/*  miniboxtk - small GUI toolkit/widget system
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    (Supported by NLnet NGI0 PET Fund in 2022)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://repo.hu/projects/miniboxtk
      contact lead developer: http://www.repo.hu/projects/miniboxtk/contact.html
*/

#include <stdlib.h>
#include <string.h>
#include "window.h"
#include "style.h"
#include "widget.h"
#include "draw.h"
#include "mpm.h"
#include <libmbtk/widget/box.h>
#include <libmbtk/widget/label.h>

#include "style_f3d.h"

typedef struct styleprop_3d_s {
	mbtk_styleprop_hdr_t hdr;

	mbtk_stroke_t frame_bright, frame_dark, frame_shadow, frame_form;
	mbtk_color_t bg, bg_activated, bg_form, fg, hilight, focus;
	signed char text_halign, text_valign;
} styleprop_3d_t;


static struct {
	mbtk_style_t style;
	styleprop_3d_t defprop; /* default style properties */
	int inited;

	/* keywords */
	struct {
		mbtk_kw_t hbox, vbox, label, checkbox, radiobutton;
		mbtk_kw_t frame_bright_color, frame_bright_width, frame_dark_color;
		mbtk_kw_t frame_dark_width, frame_shadow_color, frame_shadow_width;
		mbtk_kw_t frame_form_width, frame_form_color, bg_form_color, focus_color;
		mbtk_kw_t arrow_down, mark_check, radio_frame, radio_focus, radio_mark;
		mbtk_kw_t textedit, textedit_vbox;
	} kw;
} f3d;

static const char *mpm_mark_check[] = {
	"! MPM1",
	"10 10 3",
	"o $fg.color",
	"+ $fg.color alpha * 0.5",
	". #00000000",
	"..........",
	"..........",
	"......+o .",
	"......o+..",
	".....+o...",
	"..o..o+...",
	"..+o+o....",
	"...+o+....",
	"....+.....",
	"..........",
};

#define RADIO_SIZE 13

static const char *mpm_radio_frame[] = {
	"! MPM1",
	"13 13 5",
	". #00000000",
	", #0D0F0CFF",
	"+ #4B4B49FF",
	"@ #9B9D9AFF",
	"# $bg_form.color",
	"...@+,,,+@...",
	"..+,@###@,+..",
	".++#######++.",
	"@,#########,@",
	"+@#########@+",
	",###########,",
	",###########,",
	",###########,",
	"+@#########@+",
	"@,#########,@",
	".++#######++.",
	"..+,@###@,+..",
	"...@+,,,+@..."
};

static const char *mpm_radio_mark[] = {
	"! MPM1",
	"13 13 3",
	". #00000000",
	"# $fg.color",
	"+ $fg.color alpha * 0.5",
	".............",
	".............",
	".............",
	".............",
	".....+#+.....",
	"....+###+....",
	"....#####....",
	"....+###+....",
	".....+#+.....",
	".............",
	".............",
	".............",
	"............."
};

static const char *mpm_radio_focus[] = {
	"! MPM1",
	"13 13 2",
	". #00000000",
	"+ $focus.color",
	".............",
	".............",
	"....+++++....",
	"...+.....+...",
	"..+.......+..",
	"..+.......+..",
	"..+.......+..",
	"..+.......+..",
	"..+.......+..",
	"...+.....+...",
	"....+++++....",
	".............",
	"............."
};


static const char *mpm_arrow_down[] = {
	"! MPM1",
	"7 7 3",
	"o $fg.color",
	"+ $fg.color alpha * 0.5",
	". #00000000",
	".......",
	".......",
	"+ooooo+",
	".+ooo+.",
	"..+o+..",
	"...+...",
	"......."
};



static void f3d_draw_box_frame(mbtk_window_t *win, const mbtk_styleprop_t *sp_, mbtk_widget_flags_t flg, long x1, long y1, long wi, long he)
{
	const styleprop_3d_t *sp = (styleprop_3d_t *)sp_;
	int n, x2 = x1 + wi - 1, y2 = y1 + he - 1;

	if (flg.activated) {
		int dw = sp->frame_dark.width;

		mbtk_draw_color(win, sp->frame_dark.color.r, sp->frame_dark.color.g, sp->frame_dark.color.b, sp->frame_dark.color.a);
		for(n = 0; n < dw; n++) {
			mbtk_draw_hline(win, x1+n, y1+n, x2-1);
			mbtk_draw_vline(win, x1+n, y1+n, y2-1);
		}

		mbtk_draw_color(win, sp->frame_shadow.color.r, sp->frame_shadow.color.g, sp->frame_shadow.color.b, sp->frame_shadow.color.a);
		for(n = 0; n < sp->frame_shadow.width; n++) {
			mbtk_draw_hline(win, x1+n+dw, y1+n+dw, x2-1);
			mbtk_draw_vline(win, x1+n+dw, y1+n+dw, y2-1);
		}

		mbtk_draw_color(win, sp->frame_bright.color.r, sp->frame_bright.color.g, sp->frame_bright.color.b, sp->frame_bright.color.a);
		for(n = 0; n < sp->frame_bright.width; n++) {
			mbtk_draw_hline(win, x2-n, y2-n, x1);
			mbtk_draw_vline(win, x2-n, y2-n, y1);
		}
	}
	else {
		int sw = sp->frame_shadow.width;
		mbtk_draw_color(win, sp->frame_shadow.color.r, sp->frame_shadow.color.g, sp->frame_shadow.color.b, sp->frame_shadow.color.a);
		for(n = 0; n < sw; n++) {
			mbtk_draw_hline(win, x2-n, y2-n, x1);
			mbtk_draw_vline(win, x2-n, y2-n, y1);
		}

		mbtk_draw_color(win, sp->frame_dark.color.r, sp->frame_dark.color.g, sp->frame_dark.color.b, sp->frame_dark.color.a);
		for(n = 0; n < sp->frame_dark.width; n++) {
			mbtk_draw_hline(win, x2-n-sw, y2-n-sw, x1+n+1);
			mbtk_draw_vline(win, x2-n-sw, y2-n-sw, y1);
		}

		mbtk_draw_color(win, sp->frame_bright.color.r, sp->frame_bright.color.g, sp->frame_bright.color.b, sp->frame_bright.color.a);
		for(n = 0; n < sp->frame_bright.width; n++) {
			mbtk_draw_hline(win, x1+n, y1+n, x2-1);
			mbtk_draw_vline(win, x1+n, y1+n, y2-1-n);
		}
	}
}

static void f3d_draw_box_overlay(mbtk_window_t *win, const mbtk_styleprop_t *sp_, mbtk_widget_flags_t flg, long x1, long y1, long wi, long he)
{
	const styleprop_3d_t *sp = (styleprop_3d_t *)sp_;
	long x2 = x1 + wi - 1, y2 = y1 + he - 1;
	long ol, or, ot, ob;

	if (flg.focus) {
		if (flg.activated) {
			ol = ot = sp->frame_dark.width + sp->frame_shadow.width;
			or = ob = sp->frame_bright.width;
		}
		else {
			or = ob = sp->frame_dark.width + sp->frame_shadow.width;
			ol = ot = sp->frame_bright.width;
		}

		mbtk_draw_color(win, sp->focus.r, sp->focus.g, sp->focus.b, sp->focus.a);
		mbtk_draw_hline(win, x1+ol+1, y1+ot+1, x2-or-1);
		mbtk_draw_hline(win, x1+ol+1, y2-ob-1, x2-or-1);
		mbtk_draw_vline(win, x1+ol+1, y1+ot+1, y2-ob-1);
		mbtk_draw_vline(win, x2-or-1, y1+ot+1, y2-ob-1);
	}
}


static void f3d_draw_bg_rect(mbtk_window_t *win, const mbtk_styleprop_t *sp_, mbtk_widget_flags_t flg, long x, long y, long w, long h)
{
	const styleprop_3d_t *sp = (styleprop_3d_t *)sp_;
	if (flg.activated)
		mbtk_draw_color(win, sp->bg_activated.r, sp->bg_activated.g, sp->bg_activated.b, sp->bg_activated.a);
	else
		mbtk_draw_color(win, sp->bg.r, sp->bg.g, sp->bg.b, sp->bg.a);
	mbtk_draw_fillrect(win, x, y, x + w - 1, y + h - 1);
}

static void f3d_min_box_frame_size(mbtk_window_t *win, const mbtk_styleprop_t *sp_, mbtk_widget_flags_t flg, long *left, long *right, long *top, long *bottom)
{
	const styleprop_3d_t *sp = (styleprop_3d_t *)sp_;

	if (flg.activated) {
		*left = sp->frame_bright.width + sp->frame_shadow.width;
		*right = sp->frame_dark.width;
		*top = sp->frame_bright.width + sp->frame_shadow.width;
		*bottom= sp->frame_dark.width;
	}
	else {
		*left = sp->frame_bright.width;
		*right = sp->frame_dark.width + sp->frame_shadow.width;
		*top = sp->frame_bright.width;
		*bottom= sp->frame_dark.width + sp->frame_shadow.width;
	}
}

static void f3d_set_fg_color(mbtk_window_t *win, const mbtk_styleprop_t *sp_, mbtk_widget_flags_t flg)
{
	const styleprop_3d_t *sp = (styleprop_3d_t *)sp_;
	mbtk_draw_color(win, sp->fg.r, sp->fg.g, sp->fg.b, sp->fg.a);
}

/* Draw the background and frame for an input field ("form"), where user input
   is expected. This is typically light backround and thin dark frame. */
static void draw_form_box(mbtk_window_t *win, const mbtk_styleprop_t *sp_, mbtk_widget_flags_t flg, long x1, long y1, long w, long h, int bg, int frame)
{
	styleprop_3d_t *sp = (styleprop_3d_t *)sp_;
	int n;

	if (bg) {
		mbtk_draw_color(win, sp->bg_form.r, sp->bg_form.g, sp->bg_form.b, sp->bg_form.a);
		mbtk_draw_fillrect(win, x1, y1, x1 + w - 1, y1 + h - 1);
	}

	if (frame) {
		mbtk_draw_color(win, sp->frame_form.color.r, sp->frame_form.color.g, sp->frame_form.color.b, sp->frame_form.color.a);
		for(n = 0; n < sp->frame_form.width; n++) {
			mbtk_draw_hline(win, x1, y1+n, x1+w-1);
			mbtk_draw_hline(win, x1, y1+h-1-n, x1+w-1);
			mbtk_draw_vline(win, x1+n, y1, y1+h-1);
			mbtk_draw_vline(win, x1+w-1-n, y1, y1+h-1);
		}
	}
}

/* Draw the standard focus overlay in an input form widget */
static void draw_form_focus(mbtk_window_t *win, const mbtk_styleprop_t *sp_, mbtk_widget_flags_t flg, long x1, long y1, long w, long h)
{
	styleprop_3d_t *sp = (styleprop_3d_t *)sp_;
	int offs;

	if (!flg.focus)
		return;

	offs = sp->frame_form.width + 1;

	mbtk_draw_color(win, sp->focus.r, sp->focus.g, sp->focus.b, sp->focus.a);
	mbtk_draw_hline(win, x1+offs, y1+offs, x1+w-1-offs);
	mbtk_draw_hline(win, x1+offs, y1+h-1-offs, x1+w-1-offs);
	mbtk_draw_vline(win, x1+offs, y1+offs, y1+h-1-offs);
	mbtk_draw_vline(win, x1+w-1-offs, y1+offs, y1+h-1-offs);
}

/*** API ***/
static int f3d_getset_prop(const mbtk_style_t *style, mbtk_styleprop_t *dst_, mbtk_kw_t key, mbtk_arg_t *get, mbtk_arg_t *set)
{
	styleprop_3d_t *dst = (styleprop_3d_t *)dst_;

	/* standard style props */
	if (key == mbtk_kw_frame_width) /* useful for set only */
		return mbtk_arg_getset_uchar(&dst->frame_bright.width, NULL, set) | mbtk_arg_getset_uchar(&dst->frame_dark.width, NULL, set);

	if (key == mbtk_kw_bg_color)
		return mbtk_arg_getset_color(&dst->bg, get, set);

	if (key == mbtk_kw_fg_color)
		return mbtk_arg_getset_color(&dst->fg, get, set);

	if (key == mbtk_kw_hilight_color)
		return mbtk_arg_getset_color(&dst->hilight, get, set);

	if (key == mbtk_kw_text_halign)
		return mbtk_arg_getset_schar(&dst->text_halign, get, set);

	if (key == mbtk_kw_text_valign)
		return mbtk_arg_getset_schar(&dst->text_valign, get, set);

	/* non-standard */
	if (key == f3d.kw.focus_color)
		return mbtk_arg_getset_color(&dst->focus, get, set);

	if (key == f3d.kw.bg_form_color)
		return mbtk_arg_getset_color(&dst->bg_form, get, set);

	if (key == f3d.kw.frame_bright_color)
		return mbtk_arg_getset_color(&dst->frame_bright.color, get, set);

	if (key == f3d.kw.frame_bright_width)
		return mbtk_arg_getset_uchar(&dst->frame_bright.width, get, set);

	if (key == f3d.kw.frame_dark_color)
		return mbtk_arg_getset_color(&dst->frame_dark.color, get, set);

	if (key == f3d.kw.frame_dark_width)
		return mbtk_arg_getset_uchar(&dst->frame_dark.width, get, set);

	if (key == f3d.kw.frame_shadow_color)
		return mbtk_arg_getset_color(&dst->frame_shadow.color, get, set);

	if (key == f3d.kw.frame_shadow_width)
		return mbtk_arg_getset_uchar(&dst->frame_shadow.width, get, set);

	if (key == f3d.kw.frame_form_color)
		return mbtk_arg_getset_color(&dst->frame_form.color, get, set);

	if (key == f3d.kw.frame_form_width)
		return mbtk_arg_getset_uchar(&dst->frame_form.width, get, set);

	return -1;
}

static int f3d_minsize(const mbtk_style_t *style, mbtk_widget_t *w, long *mx, long *my)
{
	styleprop_3d_t *sp = (styleprop_3d_t *)mbtk_styleprop_get(w);
	if (w->type == f3d.kw.checkbox) {
		*mx = *my = 10 + 2 * sp->frame_form.width;
		return 0;
	}
	if (w->type == f3d.kw.radiobutton) {
		*mx = *my = RADIO_SIZE;
		return 0;
	}
	return -1;
}

static int f3d_draw_widget(const mbtk_style_t *style, mbtk_widget_t *w)
{
	mbtk_styleprop_t *sp_ = mbtk_styleprop_get(w);
	styleprop_3d_t *sp = (styleprop_3d_t *)sp_;
	mbtk_window_t *win = w->win;

	if (w->type == f3d.kw.checkbox) {
		int frw = sp->frame_form.width, size = 10 + 2 * frw;
		long ox = (w->hvbox.w - size) / 2, oy = (w->hvbox.h - size) / 2;
		mbtk_px_id_t pxid = mbtk_mpm_get(win, style, sp_, f3d.kw.mark_check, NULL);
		draw_form_box(win, sp_, w->flags, w->hvbox.x + ox, w->hvbox.y + oy, size, size, 1, 1);

		if (w->flags.activated)
			mbtk_draw_pixmap(win, w->hvbox.x + ox + frw, w->hvbox.y + oy + frw, pxid);
		return 0;
	}

	if (w->type == f3d.kw.radiobutton) {
		mbtk_px_id_t pxid_frame = mbtk_mpm_get(win, style, sp_, f3d.kw.radio_frame, NULL);
		long size = 13, ox = (w->hvbox.w - size) / 2, oy = (w->hvbox.h - size) / 2;

		mbtk_draw_pixmap(win, w->hvbox.x + ox, w->hvbox.y + oy, pxid_frame);
		if (w->flags.activated) {
			mbtk_px_id_t pxid_mark = mbtk_mpm_get(win, style, sp_, f3d.kw.radio_mark, NULL);
			mbtk_draw_pixmap(win, w->hvbox.x + ox, w->hvbox.y + oy, pxid_mark);
		}
		return 0;
	}

	/* text editor: don't draw enclosing widget frame */
	if (w->type == f3d.kw.textedit) {
		mbtk_textedit_fallback_draw_no_frame(w);
		return 0;
	}

	/* text editor: draw form-edit background on scrolled vbox (child widget) */
	if (w->purpose == f3d.kw.textedit_vbox) {
		draw_form_box(win, sp_, w->flags, w->hvbox.x, w->hvbox.y, w->hvbox.w, w->hvbox.h, 1, 0);
		return 0;
	}

	return -1;
}

static int f3d_draw_widget_overlay(const mbtk_style_t *style, mbtk_widget_t *w)
{
	mbtk_styleprop_t *sp_ = mbtk_styleprop_get(w);
	styleprop_3d_t *sp = (styleprop_3d_t *)sp_;
	mbtk_window_t *win = w->win;

	if (w->type == f3d.kw.checkbox) {
		int frw = sp->frame_form.width, size = 10 + 2 * frw;
		long ox = (w->hvbox.w - size) / 2, oy = (w->hvbox.h - size) / 2;
		draw_form_focus(win, sp_, w->flags, w->hvbox.x + ox, w->hvbox.y + oy, size, size);
		return 0;
	}

	if (w->type == f3d.kw.radiobutton) {
		mbtk_px_id_t pxid_focus = mbtk_mpm_get(win, style, sp_, f3d.kw.radio_focus, NULL);
		long size = 13, ox = (w->hvbox.w - size) / 2, oy = (w->hvbox.h - size) / 2;

		if (w->flags.focus)
			mbtk_draw_pixmap(win, w->hvbox.x + ox, w->hvbox.y + oy, pxid_focus);
		return 0;
	}

	/* text editor: don't draw focus frame in text editor (it's enough
	   to indicate focus by having the cursor) */
	if (w->purpose == f3d.kw.textedit_vbox)
		return 0;

	return -1;
}

void mbtk_f3d_style(mbtk_display_t *disp)
{
	if (!f3d.inited) {
		f3d.style.propsize = sizeof(styleprop_3d_t);
		f3d.style.getset_prop = f3d_getset_prop;
		f3d.style.minsize = f3d_minsize;
		f3d.style.draw_widget = f3d_draw_widget;
		f3d.style.draw_widget_overlay = f3d_draw_widget_overlay;
		f3d.style.draw_bg_rect = f3d_draw_bg_rect;
		f3d.style.draw_box_frame = f3d_draw_box_frame;
		f3d.style.draw_box_overlay = f3d_draw_box_overlay;
		f3d.style.set_fg_color = f3d_set_fg_color;
		f3d.style.min_box_frame_size = f3d_min_box_frame_size;
		f3d.inited = 1;
		f3d.kw.hbox  = mbtk_kw("hbox");
		f3d.kw.vbox  = mbtk_kw("vbox");
		f3d.kw.label = mbtk_kw("label");
		f3d.kw.checkbox = mbtk_kw("checkbox");
		f3d.kw.radiobutton = mbtk_kw("radiobutton");

		/* non-standard styleprops: */
		f3d.kw.frame_bright_color = mbtk_kw("frame_bright.color");
		f3d.kw.frame_bright_width = mbtk_kw("frame_bright.width");
		f3d.kw.frame_dark_color   = mbtk_kw("frame_dark.color");
		f3d.kw.frame_dark_width   = mbtk_kw("frame_dark.width");
		f3d.kw.frame_shadow_color = mbtk_kw("frame_shadow.color");
		f3d.kw.frame_shadow_width = mbtk_kw("frame_shadow.width");
		f3d.kw.frame_form_color   = mbtk_kw("frame_form.color");
		f3d.kw.frame_form_width   = mbtk_kw("frame_form.width");
		f3d.kw.bg_form_color      = mbtk_kw("bg_form.color");
		f3d.kw.focus_color        = mbtk_kw("focus.color");
		f3d.kw.mark_check         = mbtk_kw("mark_check");
		f3d.kw.arrow_down         = mbtk_kw("arrow_down");
		f3d.kw.radio_frame        = mbtk_kw("radio_frame");
		f3d.kw.radio_focus        = mbtk_kw("radio_focus");
		f3d.kw.radio_mark         = mbtk_kw("radio_mark");
		f3d.kw.textedit           = mbtk_kw("textedit");
		f3d.kw.textedit_vbox      = mbtk_kw("textedit.vbox");

		f3d.defprop.frame_bright.color = mbtk_lcolor(0xffffffff); f3d.defprop.frame_bright.width = 1;
		f3d.defprop.frame_dark.color = mbtk_lcolor(0x9a9a9aff); f3d.defprop.frame_dark.width = 1;
		f3d.defprop.frame_shadow.color = mbtk_lcolor(0x000000ff); f3d.defprop.frame_shadow.width = 1;
		f3d.defprop.frame_form.color = mbtk_lcolor(0x000000ff); f3d.defprop.frame_form.width = 1;
		f3d.defprop.bg = mbtk_lcolor(0xdbdbdbff);
		f3d.defprop.bg_activated = mbtk_lcolor(0xc8c8b8ff);
		f3d.defprop.bg_form = mbtk_lcolor(0xf0f0f0ff);
		f3d.defprop.fg = mbtk_lcolor(0x000000ff);
		f3d.defprop.hilight = mbtk_lcolor(0x0000ccff);
		f3d.defprop.focus = mbtk_lcolor(0x777700a0);
	}
	disp->style = &f3d.style;
	disp->styleprop = malloc(sizeof(styleprop_3d_t));
	memcpy(disp->styleprop, &f3d.defprop, sizeof(styleprop_3d_t));

	/* install mpms */
	mbtk_mpm_set(disp, f3d.kw.arrow_down, mpm_arrow_down);
	mbtk_mpm_set(disp, f3d.kw.mark_check, mpm_mark_check);
	mbtk_mpm_set(disp, f3d.kw.radio_frame, mpm_radio_frame);
	mbtk_mpm_set(disp, f3d.kw.radio_focus, mpm_radio_focus);
	mbtk_mpm_set(disp, f3d.kw.radio_mark, mpm_radio_mark);
}
