/*  miniboxtk - small GUI toolkit/widget system
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    (Supported by NLnet NGI0 PET Fund in 2022)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://repo.hu/projects/miniboxtk
      contact lead developer: http://www.repo.hu/projects/miniboxtk/contact.html
*/

#define MBTK_NO_INCLUDE_WIDGETS
#include <libmbtk/util.h>
#include <libmbtk/widget.h>
#include <libmbtk/widget_imp.h>
#include <libmbtk/widget/menu.h>
#include <libmbtk/widget/box.h>
#include <libmbtk/widget/label.h>
#include <libmbtk/widget/arrow.h>
#include <libmbtk/widget/checkbox.h>
#include <libmbtk/widget/separator.h>
#include <libmbtk/widget/hand_outclk.h>
#include <libmbtk/widget/Common.h>

/*
Design: this widget is a HBOX; each child is a menu label - normally
a label object, but can be a picture or anything else.
*/

static void mbtk_menu_imp_destroy(mbtk_widget_t *w)
{
	mbtk_menu_t *menu = (mbtk_menu_t *)w;
	mbtk_widget_t *par = NULL;

	/* notify our parent item (the menu bar typically) about the removal */
	if (menu->parent_item != NULL)
		par = (mbtk_widget_t *)menu->parent_item->hvbox.parent;
	if ((par != NULL) && (par->imp != NULL) && (par->imp->pre_child_unlink != NULL) && !par->destroying)
		par->imp->pre_child_unlink(par, w);
	mbtk_handler_hovradio_uninit(&menu->radio);
}

static void mbtk_menu_pre_child_unlink(mbtk_widget_t *w, mbtk_widget_t *child)
{
	mbtk_menu_t *parent = (mbtk_menu_t *)w;
	hvbox_t *n;
	long idx;

	for(n = parent->w.hvbox.ch_first, idx = 0; (n != NULL) && (idx < parent->submenus.used); n = n->next, idx++) {
		if (n == &child->hvbox) {
			mbtk_widget_t *submenu = parent->submenus.array[idx];

			vtp0_remove(&parent->submenus, idx, 1);

			/* if we are here because the menu is being destroyed, we'll need
			   to destroy the submenu. The widget tree recurses to menu labels
			   as they are widgets in the hbox but submenus are attachments. Can't free
			   them in imp_destroy() because too much of menu is destroyed by then */
			if ((submenu != NULL) && w->destroying)
				mbtk_widget_destroy(submenu);

			return; /* can't be more instances of the very same widget */
		}
	}
}

static const mbtk_widget_imp_t mbtk_menu_imp = {
	mbtk_menu_imp_destroy,

	mbtk_comm_frame_fallback_minsize,
	mbtk_comm_frame_fallback_draw,
	mbtk_comm_frame_fallback_draw_overlay,

	NULL,  /* get_prop */
	NULL,  /* set_prop */

	NULL, /* post_assign */
	NULL, /* pre_unassign */

	NULL, /* pre_appear */
	NULL, /* post_disappear */

	mbtk_menu_pre_child_unlink
};

static mbtk_event_handled_t mbtk_menu_closeall_cb(mbtk_widget_t *w, mbtk_kw_t id, void *user_data)
{
	mbtk_menu_t *menu = user_data;
	mbtk_menu_close_all(menu);
	return MBTK_EVENT_HANDLED;
}

mbtk_menu_t *mbtk_menu_new(mbtk_menu_t *caller_alloced)
{
	mbtk_menu_t *menu;
	mbtk_kw_t kw_menu, kw_child_invocation_finished, kw_outclk, kw_cancel;

	mbtk_kw_cache(kw_menu, "menu");
	mbtk_kw_cache(kw_child_invocation_finished, "child_invocation_finished");
	mbtk_kw_cache(kw_cancel, "cancel");

	mbtk_create_wt(menu, kw_menu, &mbtk_menu_imp, caller_alloced);
	menu->w.hvbox.type = HVBOX_VBOX;

	mbtk_handler_hovradio_init(NULL, &menu->radio); /* win == NULL is okay here: we are going to free the handler on widget destroy */

	mbtk_handler_outclk_install(&menu->w);
	mbtk_kw_cache(kw_outclk, "outclk");
	mbtk_callback_set_noarg(&menu->w, kw_outclk, mbtk_menu_closeall_cb, menu);
	mbtk_callback_set_noarg(&menu->w, kw_child_invocation_finished, mbtk_menu_closeall_cb, menu);
	mbtk_callback_set_noarg(&menu->w, kw_cancel, mbtk_menu_closeall_cb, menu);

	return menu;
}

static mbtk_event_handled_t mbtk_menu_radio_cb(mbtk_widget_t *w, mbtk_kw_t id, void *user_data)
{
	int idx;
	mbtk_menu_t *menu = user_data;
	hvbox_t *m;

	/* find which selector is activated, show correspoding submenu, hide all other submenus */
	for(m = menu->w.hvbox.ch_first, idx = 0; (m != NULL) && (idx < menu->submenus.used); m = m->next, idx++) {
		mbtk_widget_t *wm = (mbtk_widget_t *)m;
		mbtk_menu_t *submenu = menu->submenus.array[idx];
		if (submenu != NULL) {
			if (wm == w) {
				long x = wm->win->x + wm->hvbox.x + wm->hvbox.w, y = wm->win->y + wm->hvbox.y;
				submenu->parent_item = wm;
				mbtk_menu_open_dropdown(submenu, x, y);
			}
			else
				mbtk_menu_close(submenu);
		}
	}

	return MBTK_EVENT_HANDLED;
}

static void mbtk_menu_open_(mbtk_menu_t *menu, long screen_x, long screen_y, mbtk_window_type_t wty, mbtk_window_t *parwin)
{
	mbtk_window_t *win;

	if (menu->w.win != NULL)
		return;

	if (parwin == NULL)
		parwin = menu->parent_item->win;
	win = mbtk_window_new(parwin->display, wty, "menu", screen_x, screen_y, 32, 32);
	mbtk_window_transient4(win, parwin);

	mbtk_window_set_root_widget(win, &menu->w);

	/* auto-activate the first item */
	mbtk_hand_hovradio_activated((mbtk_widget_t *)menu->w.hvbox.ch_first);
}

void mbtk_menu_open_dropdown(mbtk_menu_t *menu, long screen_x, long screen_y)
{
	mbtk_menu_open_(menu, screen_x, screen_y, MBTK_WNTY_DROPDOWN_MENU, NULL);
}

void mbtk_menu_open_popup(mbtk_menu_t *menu, mbtk_window_t *parwin, long screen_x, long screen_y)
{
	mbtk_menu_open_(menu, screen_x, screen_y, MBTK_WNTY_POPUP_MENU, parwin);
}



void mbtk_menu_close(mbtk_menu_t *menu)
{
	mbtk_window_t *win = menu->w.win;
	int idx;
	hvbox_t *m;

	if (win == NULL)
		return;


	/* close all submenus first */
	for(m = menu->w.hvbox.ch_first, idx = 0; (m != NULL) && (idx < menu->submenus.used); m = m->next, idx++) {
		mbtk_menu_t *submenu = menu->submenus.array[idx];
		if (submenu != NULL)
			mbtk_menu_close(submenu);
	}

	mbtk_widget_unlink(&menu->w); /* menu widget tree is floating so won't get destroyed */
	mbtk_window_close(win);
}

void mbtk_menu_close_all(mbtk_menu_t *menu)
{
	mbtk_widget_t *parent = NULL;

	if (menu->parent_item != NULL)
		parent = (mbtk_widget_t *)menu->parent_item->hvbox.parent;

	mbtk_menu_close(menu);

	if (parent != NULL) {
		mbtk_kw_t kw_cancel;

		/* tell parent to close */
		mbtk_kw_cache(kw_cancel, "cancel");
		if (parent != NULL)
			mbtk_callback_noarg(parent, kw_cancel);
	}
}

MBTK_INLINE void mbtk_menu_post_add(mbtk_menu_t *menu, mbtk_widget_t *row)
{
	mbtk_kw_t kw_toggled;

	row->flags.activated = 0;

	mbtk_widget_embedded((mbtk_widget_t *)row->hvbox.parent, row);

	mbtk_handler_hovradio_add_widget(&menu->radio, row);
	mbtk_kw_cache(kw_toggled, "toggled");
	mbtk_callback_set_noarg(row, kw_toggled, mbtk_menu_radio_cb, menu);
}

void mbtk_menu_append(mbtk_menu_t *menu, mbtk_widget_t *row)
{
	mbtk_box_add_widget((mbtk_box_t *)menu, row, 0);
	mbtk_menu_post_add(menu, row);
}

void mbtk_menu_add_at(mbtk_widget_t *existing_row, mbtk_widget_t *new_row, int ins_before)
{
	mbtk_menu_t *menu = (mbtk_menu_t *)existing_row->hvbox.parent;
	mbtk_box_add_widget_at(existing_row, new_row, ins_before);
	mbtk_menu_post_add(menu, new_row);
}


mbtk_box_t *mbtk_menu_build_stdrow(mbtk_widget_t *prefix, mbtk_widget_t *entry, mbtk_widget_t *suffix)
{
	mbtk_box_t *row = mbtk_hbox_new(NULL);

	if (prefix == NULL) {
		mbtk_box_t *box = mbtk_vbox_new(NULL);
		box->no_background = 1;
		prefix = (mbtk_widget_t *)box;
	}

	if (suffix == NULL) {
		mbtk_box_t *box = mbtk_vbox_new(NULL);
		box->no_background = 1;
		suffix = (mbtk_widget_t *)box;
	}

	mbtk_widget_set_minsize(prefix, 20, 0);
	mbtk_widget_set_span(entry, HVBOX_FILL);
	mbtk_widget_set_minsize(suffix, 64, 0);

	mbtk_box_add_widget(row, prefix, 0);
	mbtk_box_add_widget(row, entry, 0);
	mbtk_box_add_widget(row, suffix, 0);

	return row;
}

mbtk_box_t *mbtk_menu_append_stdrow(mbtk_menu_t *menu, mbtk_widget_t *prefix, mbtk_widget_t *entry, mbtk_widget_t *suffix)
{
	mbtk_box_t *row = mbtk_menu_build_stdrow(prefix, entry, suffix);
	mbtk_menu_append(menu, &row->w);

	return row;
}

mbtk_box_t *mbtk_menu_append_label(mbtk_menu_t *menu, const char *menu_name)
{
	mbtk_label_t *lab = mbtk_label_new(NULL, menu_name);

	if (lab == NULL)
		return NULL;

	lab->halign = -1;
	return mbtk_menu_append_stdrow(menu, NULL, &lab->w, NULL);
}

mbtk_box_t *mbtk_menu_build_label_full_stdrow(int chkbox, const char *item_name, const char *hotkey)
{
	mbtk_label_t *lab = mbtk_label_new(NULL, item_name), *suffix = NULL;
	mbtk_widget_t *prefix = NULL;

	if (lab == NULL)
		return NULL;

	lab->halign = -1;

	if (chkbox) {
		mbtk_checkbox_t *chk = mbtk_checkbox_new(NULL);
		chk->w.handler = NULL; /* the application should be maintaining this */
		prefix = &chk->w;
	}

	if (hotkey != NULL) {
		suffix = mbtk_label_new(NULL, hotkey);
		if (suffix != NULL)
			suffix->halign = +1;
	}

	return mbtk_menu_build_stdrow(prefix, &lab->w, (mbtk_widget_t *)suffix);
}


mbtk_box_t *mbtk_menu_append_label_full(mbtk_menu_t *menu, int chkbox, const char *item_name, const char *hotkey)
{
	mbtk_box_t *row = mbtk_menu_build_label_full_stdrow(chkbox, item_name, hotkey);

	if (row != NULL)
		mbtk_menu_append(menu, &row->w);

	return row;
}

mbtk_widget_t *mbtk_menu_append_sep(mbtk_menu_t *menu)
{
	mbtk_separator_t *sep = mbtk_hsep_new(NULL);

	if (sep == NULL)
		return NULL;

	mbtk_menu_append(menu, &sep->w);
	return &sep->w;
}


static long submenu_idx(mbtk_widget_t *menu_row, mbtk_menu_t **parout)
{
	mbtk_menu_t *parent;
	hvbox_t *n;
	mbtk_kw_t kw_menu;
	long idx;

	mbtk_kw_cache(kw_menu, "menu");

	parent = (mbtk_menu_t *)menu_row->hvbox.parent;
	if (parent->w.type != kw_menu)
		return -1;

	*parout = parent;

	for(n = parent->w.hvbox.ch_first, idx = 0; n != NULL; n = n->next, idx++)
		if (n == &menu_row->hvbox)
			return idx;

	return -1;
}

mbtk_box_t *mbtk_menu_build_label_submenu_stdrow(const char *item_name)
{
	mbtk_box_t *res;
	mbtk_label_t *lab = mbtk_label_new(NULL, item_name);
	mbtk_arrow_t *arrow = mbtk_arrow_new(NULL, MBTK_ARROW_RIGHT);

	if ((lab == NULL) || (arrow == NULL)) {
		if (lab != NULL) mbtk_widget_destroy(&lab->w);
		if (arrow != NULL) mbtk_widget_destroy(&arrow->w);
		return NULL;
	}

	lab->halign = -1;

	res = mbtk_menu_build_stdrow(NULL, &lab->w, &arrow->w);

	mbtk_widget_set_minsize(&arrow->w, 1, 1); /* for proper right align */

	return res;
}

mbtk_box_t *mbtk_menu_append_label_submenu(mbtk_menu_t *menu, const char *item_name)
{
	mbtk_box_t *res = mbtk_menu_build_label_submenu_stdrow(item_name);
	mbtk_menu_append(menu, &res->w);
	return res;
}

int mbtk_menu_item_attach_submenu(mbtk_widget_t *menu_item, mbtk_menu_t *submenu)
{
	mbtk_menu_t *menu;
	long idx = submenu_idx(menu_item, &menu);

	if (idx < 0)
		return -1;

	vtp0_set(&menu->submenus, idx, submenu);
	return 0;
}


void mbtk_menu_item_callback_invoked(mbtk_widget_t *row, mbtk_event_handled_t (*cb)(mbtk_widget_t *w, mbtk_kw_t id, void *user_data), void *user_data)
{
	mbtk_kw_t kw_invoked;
	mbtk_kw_cache(kw_invoked, "invoked");
	mbtk_callback_set_noarg(row, kw_invoked, cb, user_data);
}

void mbtk_menu_stdrow_callback_invoked(mbtk_box_t *row, mbtk_event_handled_t (*cb)(mbtk_widget_t *w, mbtk_kw_t id, void *user_data), void *user_data)
{
	mbtk_menu_item_callback_invoked(&row->w, cb, user_data);
}

static mbtk_checkbox_t *menu_get_checkbox(mbtk_box_t *row)
{
	mbtk_checkbox_t *chk = (mbtk_checkbox_t *)row->w.hvbox.ch_first;
	mbtk_kw_t kw_checkbox;

	mbtk_kw_cache(kw_checkbox, "checkbox");

	if ((chk == NULL) || (chk->w.type != kw_checkbox))
		return NULL;

	return chk;
}

int mbtk_menu_stdrow_checkbox_get(mbtk_box_t *row)
{
	mbtk_checkbox_t *chk = menu_get_checkbox(row);

	if (chk == NULL) {
		mbtk_error("mbtk_menu_stdrow_checkbox_get(): checkbox not found\n");
		return -1;
	}

	return chk->w.flags.activated;
}

void mbtk_menu_stdrow_checkbox_set(mbtk_box_t *row, int checked)
{
	mbtk_checkbox_t *chk = menu_get_checkbox(row);

	if (chk == NULL) {
		mbtk_error("mbtk_menu_stdrow_checkbox_set(): checkbox not found\n");
		return;
	}

	chk->w.flags.activated = checked;
	mbtk_widget_needs_redraw(&chk->w);
}

int mbtk_menu_stdrow_checkbox_toggle(mbtk_box_t *row)
{
	mbtk_checkbox_t *chk = menu_get_checkbox(row);

	if (chk == NULL) {
		mbtk_error("mbtk_menu_stdrow_checkbox_toggle(): checkbox not found\n");
		return -1;
	}

	chk->w.flags.activated = !chk->w.flags.activated;
	mbtk_widget_needs_redraw(&chk->w);

	return chk->w.flags.activated;
}


