/*  miniboxtk - small GUI toolkit/widget system
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    (Supported by NLnet NGI0 PET Fund in 2022)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://repo.hu/projects/miniboxtk
      contact lead developer: http://www.repo.hu/projects/miniboxtk/contact.html
*/

#define MBTK_NO_INCLUDE_WIDGETS
#include <libmbtk/widget.h>
#include <libmbtk/widget_imp.h>
#include <libmbtk/widget/menu_bar.h>
#include <libmbtk/widget/box.h>
#include <libmbtk/widget/label.h>
#include <libmbtk/widget/Common.h>
#include <libmbtk/util.h>

/*
Design: this widget is a HBOX; each child is a menu label - normally
a label object, but can be a picture or anything else.
*/

static void mbtk_menu_bar_imp_destroy(mbtk_widget_t *w)
{
	mbtk_menu_bar_t *menu_bar = (mbtk_menu_bar_t *)w;
	mbtk_handler_hovradio_uninit(&menu_bar->radio);
}

static void mbtk_menu_bar_pre_child_unlink(mbtk_widget_t *w, mbtk_widget_t *child)
{
	mbtk_menu_bar_t *parent = (mbtk_menu_bar_t *)w;
	hvbox_t *n;
	long idx;

	/* if the label widget is removed */
	for(n = parent->w.hvbox.ch_first, idx = 0; n != NULL; n = n->next, idx++) {
		if (n == &child->hvbox) {
			mbtk_widget_t *menu = parent->menus.array[idx];

			vtp0_remove(&parent->menus, idx, 1);

			/* if we are here because the menu bar is being destroyed, we'll need
			   to destroy the menu. The widget tree recurses to menu_bar labels
			   as they are widgets in the hbox but menus are attachments. Can't free
			   them in imp_destroy() because too much of menu_bar is destroyed by then */
			if ((menu != NULL) && w->destroying)
				mbtk_widget_destroy(menu);
			return; /* can't be more instances of the very same widget */
		}
	}

	/* if the menu removed */
	for(n = parent->w.hvbox.ch_first, idx = 0; idx < parent->menus.used; n = n->next, idx++) {
		if (parent->menus.array[idx] == child) {
			mbtk_widget_destroy((mbtk_widget_t *)n); /* this calls back this very function removing the label */
			return; /* can't be more instances of the very same menu */
		}
	}
}


static const mbtk_widget_imp_t mbtk_menu_bar_imp = {
	mbtk_menu_bar_imp_destroy,

	mbtk_comm_frame_fallback_minsize,
	mbtk_comm_frame_fallback_draw,
	mbtk_comm_frame_fallback_draw_overlay,

	NULL,  /* get_prop */
	NULL,  /* set_prop */

	NULL, /* post_assign */
	NULL, /* pre_unassign */

	NULL, /* pre_appear */
	NULL, /* post_disappear */

	mbtk_menu_bar_pre_child_unlink
};

/* called back on click-away or menu item invocation, deactivates the menu bar */
static mbtk_event_handled_t mbtk_menu_bar_cancel_cb(mbtk_widget_t *w, mbtk_kw_t id, void *user_data)
{
	mbtk_menu_bar_t *menu_bar = (mbtk_menu_bar_t *)w;
	mbtk_widget_t *m1 = (mbtk_widget_t *)w->hvbox.ch_first;
	hvbox_t *m;
	long idx;

	if (m1 != NULL)
		mbtk_hand_hovradio_deactivate((mbtk_handler_hovradio_t *)m1->handler, 0);

	/* close all menus */
	for(m = menu_bar->w.hvbox.ch_first, idx = 0; m != NULL; m = m->next, idx++)
		if (menu_bar->menus.array[idx] != NULL)
				mbtk_menu_close(menu_bar->menus.array[idx]);

	return MBTK_EVENT_HANDLED;
}


mbtk_menu_bar_t *mbtk_menu_bar_new(mbtk_menu_bar_t *caller_alloced)
{
	mbtk_menu_bar_t *menu_bar;
	mbtk_kw_t kw_menu_bar, kw_child_invocation_finished, kw_cancel;

	mbtk_kw_cache(kw_menu_bar, "menu_bar");
	mbtk_kw_cache(kw_child_invocation_finished, "child_invocation_finished");
	mbtk_kw_cache(kw_cancel, "cancel");

	mbtk_create_wt(menu_bar, kw_menu_bar, &mbtk_menu_bar_imp, caller_alloced);
	menu_bar->w.hvbox.type = HVBOX_HBOX;
	mbtk_widget_set_align(&menu_bar->w, HVBOX_LEFT_TOP);
	mbtk_widget_set_spacing(&menu_bar->w, 16);

	mbtk_handler_hovradio_init(NULL, &menu_bar->radio); /* win == NULL is okay here: we are going to free the handler on widget destroy */

	mbtk_callback_set_noarg(&menu_bar->w, kw_child_invocation_finished, mbtk_menu_bar_cancel_cb, NULL);
	mbtk_callback_set_noarg(&menu_bar->w, kw_cancel, mbtk_menu_bar_cancel_cb, NULL);

	return menu_bar;
}

static mbtk_event_handled_t mbtk_menu_bar_radio_cb(mbtk_widget_t *w, mbtk_kw_t id, void *user_data)
{
	int idx;
	mbtk_menu_bar_t *menu_bar = user_data;
	hvbox_t *m;

	/* find which selector is activated, show correspoding page, hide all other pages */
	for(m = menu_bar->w.hvbox.ch_first, idx = 0; m != NULL; m = m->next, idx++) {
		mbtk_widget_t *wm = (mbtk_widget_t *)m;

		if (menu_bar->menus.array[idx] != NULL) {
			if (wm->flags.activated) {
				long x = w->win->x + w->hvbox.x, y = w->win->y + w->hvbox.y + w->hvbox.h;
				mbtk_menu_open_dropdown(menu_bar->menus.array[idx], x, y);
			}
			else
				mbtk_menu_close(menu_bar->menus.array[idx]);
		}
	}

	return MBTK_EVENT_HANDLED;
}

/* called after a label ahs been added to the hbox of a menu bar */
MBTK_INLINE void mbtk_menu_bar_post_add(mbtk_menu_bar_t *menu_bar, mbtk_widget_t *label, mbtk_menu_t *menu)
{
	mbtk_kw_t kw_toggled;

	label->flags.activated = 0;

	mbtk_widget_embedded((mbtk_widget_t *)label->hvbox.parent, label);

	mbtk_handler_hovradio_add_widget(&menu_bar->radio, label);
	mbtk_kw_cache(kw_toggled, "toggled");
	mbtk_callback_set_noarg(label, kw_toggled, mbtk_menu_bar_radio_cb, menu_bar);

	if (menu != NULL)
		menu->parent_item = label;
}


mbtk_widget_t *mbtk_menu_bar_append(mbtk_menu_bar_t *menu_bar, mbtk_widget_t *label, mbtk_menu_t *menu)
{
	mbtk_box_add_widget((mbtk_box_t *)menu_bar, label, 0);
	mbtk_menu_bar_post_add(menu_bar, label, menu);
	vtp0_append(&menu_bar->menus, menu);
	return label;
}

mbtk_widget_t *mbtk_menu_bar_add_at(mbtk_widget_t *existing, mbtk_widget_t *label, mbtk_menu_t *menu, int ins_before)
{
	mbtk_kw_t kw_menu_bar;
	mbtk_menu_bar_t *menu_bar = (mbtk_menu_bar_t *)existing->hvbox.parent;
	int n;

	mbtk_kw_cache(kw_menu_bar, "menu_bar");

	if (menu_bar->w.type != kw_menu_bar) {
		mbtk_error("mbtk_menu_bar_add_at(): existing widget's parent is not a menu_bar\n");
		return NULL;
	}

	mbtk_box_add_widget_at(existing, label, ins_before);
	mbtk_menu_bar_post_add(menu_bar, label, menu);

	/* figure where to insert the menu */
	for(n = 0; n < menu_bar->menus.used; n++) {
		mbtk_menu_t *em = menu_bar->menus.array[n];
		if (em->parent_item == existing) {
			if (!ins_before)
				n++;
			vtp0_insert_len(&menu_bar->menus, n, (void **)&menu, 1);
			return label;
		}
	}

	/* existing widget not found */
	mbtk_error("mbtk_menu_bar_add_at(): internal error: confused menu vector\n");
	return NULL;
}


mbtk_widget_t *mbtk_menu_bar_append_label(mbtk_menu_bar_t *menu_bar, const char *menu_name, mbtk_menu_t *menu)
{
	mbtk_label_t *lab = mbtk_label_new(NULL, menu_name);
	lab->draw_bg = 1;
	return mbtk_menu_bar_append(menu_bar, &lab->w, menu);
}

mbtk_widget_t *mbtk_menu_bar_add_at_label(mbtk_widget_t *existing, const char *menu_name, mbtk_menu_t *menu, int ins_before)
{
	mbtk_label_t *lab = mbtk_label_new(NULL, menu_name);
	lab->draw_bg = 1;
	return mbtk_menu_bar_add_at(existing, &lab->w, menu, ins_before);
}

