/*  miniboxtk - small GUI toolkit/widget system
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    (Supported by NLnet NGI0 PET Fund in 2022)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://repo.hu/projects/miniboxtk
      contact lead developer: http://www.repo.hu/projects/miniboxtk/contact.html
*/

#define MBTK_NO_INCLUDE_WIDGETS
#include <libmbtk/widget.h>
#include <libmbtk/util.h>
#include <libmbtk/widget_imp.h>
#include <libmbtk/callback_imp.h>
#include <libmbtk/widget/box.h>
#include <libmbtk/widget/tbl.h>

static int mbtk_tbl_fallback_minsize(mbtk_widget_t *w, long *mx, long *my)
{
	mbtk_tbl_t *tbl = (mbtk_tbl_t *)w;
	const mbtk_style_t *style = mbtk_widget_style(w);
	mbtk_styleprop_t *sp = mbtk_styleprop_get(w);

	*mx = *my = 0;

	if (tbl->framed) {
		long l, r, t, b;
		style->min_box_frame_size(w->win, sp, w->flags, &l, &r, &t, &b);
		w->hvbox.fr_l = l; w->hvbox.fr_r = r;
		w->hvbox.fr_t = t; w->hvbox.fr_b = b;
		*mx += l+r; *my += t+b;
	}

	return 0;
}

static void darken(mbtk_arg_t *dst, mbtk_arg_t src, double factor)
{
	unsigned long r, g, b, a;

	if (src.type != MBTK_ULONG) {
		mbtk_error("tbl row color: invalid style: wrong color type\n");
		memset(dst, 0, sizeof(mbtk_arg_t));
		return;
	}

	r = (src.v.ul & 0xff000000ul) >> 24;
	g = (src.v.ul & 0x00ff0000ul) >> 16;
	b = (src.v.ul & 0x0000ff00ul) >> 8;
	a = (src.v.ul & 0x00000ff0ul);

	dst->type = MBTK_ULONG;

	r = (double)r * factor;
	g = (double)g * factor;
	b = (double)b * factor;

	dst->v.ul = (r << 24) | (g << 16) | (b << 8) | a;
}

static void set_row_color(mbtk_widget_t *w, const mbtk_style_t *style, mbtk_styleprop_t *sp, int alt, int first)
{
	mbtk_kw_t kw_fclr;
	mbtk_window_t *win = w->win;
	mbtk_tbl_t *tbl = (mbtk_tbl_t *)w;
	mbtk_arg_t defclr, clr;

	kw_fclr = mbtk_kw("bg_form.color");
	if (mbtk_style_get_prop(style, sp, kw_fclr, &defclr) != 0) {
		if (mbtk_style_get_prop(style, sp, mbtk_kw_bg_color, &defclr) != 0) { /* must have */
			mbtk_error("tbl row color: invalid style: no bg_color\n");
			return;
		}
	}

	if (first && tbl->bg_first) {
		mbtk_kw_t kw_tbl_hdr;
		kw_tbl_hdr = mbtk_kw("tbl.header_color");
		if (mbtk_style_get_prop(style, sp, kw_tbl_hdr, &clr) != 0)
			darken(&clr, defclr, 0.6);
		mbtk_draw_color_ul(win, clr.v.ul);
	}
	else if (alt) {
		mbtk_kw_t kw_tbl_hdr;
		kw_tbl_hdr = mbtk_kw("tbl.header_color");
		if (mbtk_style_get_prop(style, sp, kw_tbl_hdr, &clr) != 0)
			darken(&clr, defclr, 0.8);
		mbtk_draw_color_ul(win, clr.v.ul);
	}
	else
		mbtk_draw_color_ul(win, defclr.v.ul);
}

static void mbtk_tbl_fallback_draw(mbtk_widget_t *w)
{
	mbtk_tbl_t *tbl = (mbtk_tbl_t *)w;
	mbtk_window_t *win = w->win;
	const mbtk_style_t *style = mbtk_widget_style(w);
	mbtk_styleprop_t *sp = mbtk_styleprop_get(w);

	if (!tbl->no_background) {
		if (tbl->bg_even_odd) {
			long col = 0, y, y2;
			int first = 1, alt = 0;
			hvbox_t *n;

			/* per row coloring */
			y = w->hvbox.y;
			for(n = w->hvbox.ch_first; n != NULL; n = n->next) {
				if (col == 0) {
					set_row_color(w, style, sp, alt, first);
					y2 = n->y + n->h - w->hvbox.spacing/2 - 1;
					mbtk_draw_fillrect(win, w->hvbox.x, y, w->hvbox.x + w->hvbox.w - 1, y2);
					y = y2+1;
					alt = !alt;
					first = 0;
				}

				col++;
				if (col >= w->hvbox.tbl_cols)
					col = 0;
			}

			/* fill remaining space at the bottom with the last row's color */
			mbtk_draw_fillrect(win, w->hvbox.x, y, w->hvbox.x + w->hvbox.w - 1, w->hvbox.y + w->hvbox.h - 1);
		}
		else
			style->draw_bg_rect(win, sp, w->flags, w->hvbox.x, w->hvbox.y, w->hvbox.w, w->hvbox.h);
	}

	if (tbl->framed)
		style->draw_box_frame(win, sp, w->flags, w->hvbox.x, w->hvbox.y, w->hvbox.w, w->hvbox.h);
}

static void mbtk_tbl_fallback_draw_overlay(mbtk_widget_t *w)
{
	mbtk_window_t *win = w->win;
	const mbtk_style_t *style = mbtk_widget_style(w);
	mbtk_styleprop_t *sp = mbtk_styleprop_get(w);

	style->draw_box_overlay(win, sp, w->flags, w->hvbox.x, w->hvbox.y, w->hvbox.w, w->hvbox.h);
}


static const mbtk_widget_imp_t mbtk_tbl_imp = {
	NULL, /* mbtk_tbl_destroy */

	mbtk_tbl_fallback_minsize,
	mbtk_tbl_fallback_draw,
	mbtk_tbl_fallback_draw_overlay,

	NULL,  /* get_prop */
	NULL,  /* set_prop */

	NULL, /* post_assign */
	NULL, /* pre_unassign */

	NULL, /* pre_appear */
	NULL, /* post_disappear */

	NULL  /* pre_child_unlink */
};

mbtk_tbl_t *mbtk_tbl_new(mbtk_tbl_t *caller_alloced, unsigned short num_cols)
{
	mbtk_tbl_t *tbl;
	mbtk_kw_t kw_tbl;

	mbtk_kw_cache(kw_tbl, "vtbl");

	if (num_cols == 0)
		return NULL;

	mbtk_create_wt(tbl, kw_tbl, &mbtk_tbl_imp, caller_alloced);
	tbl->w.hvbox.type = HVBOX_TABLE;
	tbl->w.hvbox.tbl_cols = num_cols;

	return tbl;
}

static mbtk_widget_t *mbtk_tbl_add_widget_at_(mbtk_tbl_t *parent, mbtk_widget_t *at, mbtk_widget_t *w, int ins_before)
{
	/* sanity checks */
	if (parent == NULL)
		return NULL;
	if (parent->w.hvbox.type != HVBOX_TABLE)
		return NULL;
	if (w->win != NULL)
		return NULL;

	mbtk_widget_embedded(&parent->w, w);
	mbtk_hvbox_install(w, &parent->w, at, ins_before);
	return w;
}

mbtk_widget_t *mbtk_tbl_add_widget(mbtk_tbl_t *parent, mbtk_widget_t *w, int prepend)
{
	return mbtk_tbl_add_widget_at_(parent, NULL, w, prepend);
}

void mbtk_tbl_set_framed(mbtk_tbl_t *tbl, mbtk_bool_t has_frame)
{
	if (tbl->framed != has_frame) {
		tbl->framed = has_frame;
		mbtk_widget_changed(&tbl->w, 0); /* recalculate minimum size because the frame affects it */
	}
}


void mbtk_tbl_set_homogenous_cols(mbtk_tbl_t *tbl, mbtk_bool_t hom)
{
	if (tbl->w.hvbox.tbl_hom_cols == !!hom) return;
	tbl->w.hvbox.tbl_hom_cols = !!hom;
	mbtk_widget_changed(&tbl->w, 0);
}

void mbtk_tbl_set_homogenous_rows(mbtk_tbl_t *tbl, mbtk_bool_t hom)
{
	if (tbl->w.hvbox.tbl_hom_rows == !!hom) return;
	tbl->w.hvbox.tbl_hom_rows = !!hom;
	mbtk_widget_changed(&tbl->w, 0);
}

void mbtk_tbl_set_homogenous(mbtk_tbl_t *tbl, mbtk_bool_t col, mbtk_bool_t row)
{
	if ((tbl->w.hvbox.tbl_hom_cols == !!col) && (tbl->w.hvbox.tbl_hom_rows == !!row)) return;
	tbl->w.hvbox.tbl_hom_cols = !!col;
	tbl->w.hvbox.tbl_hom_rows = !!row;
	mbtk_widget_changed(&tbl->w, 0);
}

