/*  miniboxtk - small GUI toolkit/widget system
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    (Supported by NLnet NGI0 PET Fund in 2022)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://repo.hu/projects/miniboxtk
      contact lead developer: http://www.repo.hu/projects/miniboxtk/contact.html
*/

#ifndef LIBMBTK_WIDGET_TEXTLINE_H
#define LIBMBTK_WIDGET_TEXTLINE_H

#include <libmbtk/widget.h>
#include <libmbtk/widget_inlines.h>
#include <libmbtk/src_3rd/genvector/gds_char.h>

/* A single line of a textedit. It's a special case, simplified label widget.
   Normally not directly used by application code but by composite widgets
   like textedit. */
typedef struct mbtk_textline_s {
	mbtk_widget_t w;

	/* configuration */
	mbtk_font_id_t fid;      /* font (style and size) to use for drawing the text */

	/* internal */
	int cursor;              /* the character index the cursor is before, or -1 if there's no cursor */

	unsigned width_valid:1;  /* calculated font width values */
	unsigned overwrite:1;    /* if set draw overwrite cursor, else insert cursor */

	unsigned long width;     /* in pixels */
	void *cookie;            /* backend draw */
	gds_t str;
	int font_asc, font_dsc;  /* ascend and descend of the font */
	int cursor_x, cursor_x2; /* pixel coord for ->last_cursor; x2 is used for the overwrite cursor */
} mbtk_textline_t;

/* str is always copied in these calls */
mbtk_textline_t *mbtk_textline_new(mbtk_textline_t *caller_alloced, const char *str);
void mbtk_textline_set_str(mbtk_textline_t *textline, const char *str);
void mbtk_textline_set_str_len(mbtk_textline_t *textline, const char *str, long len);

/* append str at the end of textline without changing cursor position */
int mbtk_textline_type_str(mbtk_textline_t *textline, const char *str);
void mbtk_textline_append_str_len(mbtk_textline_t *textline, const char *str, long len);

/* Remove characters from the end of the line so the line has at most len
   characters. */
void mbtk_textline_truncate(mbtk_textline_t *textline, long len);


/* Move the cursor to character position; delta values:
   -1: place insert cursor before character
    0: place overwrite cursor on character
   +1: place insert cursor after the character.
   Return 0 on success */
int mbtk_textline_cursor_set_chr(mbtk_textline_t *textline, long chpos, int delta);

/* Set cursor by window relative x coord. Return 0 on success */
int mbtk_textline_cursor_set_win(mbtk_textline_t *textline, long x);

/* Hide cursor */
void mbtk_textline_cursor_remove(mbtk_textline_t *textline);

/* Add characters to the textline at cursor (inserting or overwriting). Returns
   0 on success, -1 on error (e.g. no cursor set) */
int mbtk_textline_type_str(mbtk_textline_t *textline, const char *str);

/* Move cursor in front or at end (typically: key home and end) */
int mbtk_textline_cursor_home(mbtk_textline_t *textline);
int mbtk_textline_cursor_end(mbtk_textline_t *textline);

/* Remove num characters from the cursor (typicall: 1 for the del key) */
void mbtk_textline_del(mbtk_textline_t *textline, int num);

/* Remove a single character from before the cursor (typically for the backspace key) */
void mbtk_textline_backspace(mbtk_textline_t *textline);


MBTK_WIDGET_INLINES(textline)


/*** Internal ***/

/* Call this after directly making changes to ->str */
void mbtk_textline_str_changed(mbtk_textline_t *textline);


#endif
